##
# File:    Connection.py
# Author:  J. Westbrook
# Date:    1-Apr-2018
#
#  Connection methods for  Crate DB.
#
# Updates:
#
##
"""
Connection methods for  Crate DB.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"


import copy
import logging

from crate import client

# from crate.client.exceptions import (DatabaseError, OperationalError, ProgrammingError, Warning)

logger = logging.getLogger(__name__)


class Connection(object):

    """Class to encapsulate Crate RDBMS DBI connection."""

    def __init__(self, cfgOb=None, infoD=None, resourceName=None, sectionName="site_info_configuration", verbose=False):
        self.__verbose = verbose

        self.__db = None
        self.__dbcon = None

        self.__infoD = infoD
        self.__dbName = None

        self.__dbHost = None
        self.__dbUser = None
        self.__dbPw = None
        self.__dbSocket = None
        self.__dbPort = None
        self.__dbAdminDb = None
        self.__dbPort = None
        #
        self.__defaultPort = 4200
        self.__dbServer = "crate"
        self.__resourceName = resourceName

        self.__cfgOb = cfgOb
        #
        if infoD:
            self.setPreferences(infoD)
        #
        if resourceName:
            self.assignResource(resourceName, sectionName)

    def getPreferences(self):
        return self.__infoD

    def setPreferences(self, infoD):
        try:
            self.__infoD = copy.deepcopy(infoD)
            self.__dbName = self.__infoD.get("DB_NAME", None)
            self.__dbHost = self.__infoD.get("DB_HOST", "localhost")
            self.__dbUser = self.__infoD.get("DB_USER", None)
            self.__dbPw = self.__infoD.get("DB_PW", None)
            self.__dbSocket = self.__infoD.get("DB_SOCKET", None)
            self.__dbServer = self.__infoD.get("DB_SERVER", "crate")
            #
            port = self.__infoD.get("DB_PORT", self.__defaultPort)
            if port and str(port):
                self.__dbPort = int(str(port))
        except Exception as e:
            logger.exception("Failing with %s", str(e))

    def assignResource(self, resourceName=None, sectionName=None):
        #
        defaultPort = 4200
        defaultHost = "localhost"
        dbServer = "crate"

        self.__resourceName = resourceName
        infoD = {}
        if not self.__cfgOb:
            return infoD
        #
        if resourceName == "CRATE_DB":
            infoD["DB_NAME"] = self.__cfgOb.get("CRATE_DB_NAME", sectionName=sectionName)
            infoD["DB_HOST"] = self.__cfgOb.get("CRATE_DB_HOST", sectionName=sectionName)
            infoD["DB_SOCKET"] = self.__cfgOb.get("CRATE_DB_SOCKET", default=None, sectionName=sectionName)
            infoD["DB_PORT"] = int(str(self.__cfgOb.get("CRATE_DB_PORT", default=defaultPort, sectionName=sectionName)))
            infoD["DB_USER"] = self.__cfgOb.get("CRATE_DB_USER_NAME", sectionName=sectionName)
            infoD["DB_PW"] = self.__cfgOb.get("CRATE_DB_PASSWORD", sectionName=sectionName)

        else:
            infoD["DB_NAME"] = self.__cfgOb.get("DB_NAME", sectionName=sectionName)
            infoD["DB_HOST"] = self.__cfgOb.get("DB_HOST", default=defaultHost, sectionName=sectionName)
            infoD["DB_SOCKET"] = self.__cfgOb.get("DB_SOCKET", default=None, sectionName=sectionName)
            infoD["DB_PORT"] = int(str(self.__cfgOb.get("DB_PORT", default=defaultPort, sectionName=sectionName)))
            infoD["DB_USER"] = self.__cfgOb.get("DB_USER_NAME", sectionName=sectionName)
            infoD["DB_PW"] = self.__cfgOb.get("DB_PASSWORD", sectionName=sectionName)
        #
        infoD["DB_SERVER"] = dbServer
        self.setPreferences(infoD)
        #
        return copy.deepcopy(infoD)
        #

    def connect(self):
        """Create a database connection and return a connection object.

        Returns None on failure
        """
        #
        crateHost = "{host}:{port}".format(host=self.__dbHost, port=self.__dbPort)
        crateUri = "http://%s" % crateHost
        logger.debug("Connection using uri %s", crateUri)
        #
        dbcon = client.connect(crateUri)
        #
        if self.__dbcon is not None:
            # Close an open connection -
            logger.info("Closing an existing connection.\n")
            self.close()
        try:
            dbcon = self.__dbcon = dbcon
        except Exception as e:
            logger.exception("Connection error to server %s host %s port %d %s", self.__dbServer, self.__dbHost, self.__dbPort, str(e))
            self.__dbcon = None

        return self.__dbcon

    def close(self):
        """Close any open database connection."""
        if self.__dbcon is not None:
            try:
                self.__dbcon.close()
                self.__dbcon = None
                return True
            except Exception as e:
                logger.exception("Connection close error %s", str(e))
        return False

    def __enter__(self):
        return self.connect()

    def __exit__(self, *args):
        return self.close()



##
# File:    ContentDefinition.py
# Author:  J. Westbrook
# Date:    8-May-2018
# Version: 0.001 Initial version
#
# Updates:
#  23-May-2018 jdw add method to return vanilla data Type dictionary/revise api -
#   5-Jun-2018 jdw replace IoAdapter() with MarshalUtil() -
#  10-Jun-2018 jdw move the base table definitions in terms of content classes
#  16-Jun-2018 jdw standardize method name for getAttributeDataTypeD()
#  26-Nov-2018 jdw turn down boundary value level to standard ddl values
#   3-Feb-2019 jdw get all child items using self.__dApi.getFullDescendentList()
#  31-Mar-2019 jdw include parent details in __getAttributeFeatures()
#   6-Jun-2019 jdw take dictionary API as an argument.
#  22-Aug-2019 jdw unify naming conventions dictSubset->databaseName and *ForSubset() -> *ForDatabase()
#   5-Sep-2019 jdw add extended DDL metadata.
#
#
##
"""
Assemble configuration and dictionary metadata required to build/load database schema definitions ...

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging
import textwrap
from collections import OrderedDict

from mmcif.api.PdbxContainers import CifName

logger = logging.getLogger(__name__)


class ContentDefinition(object):
    """Assemble configuration and dictionary metadata required to build/load database schema definitions ..."""

    def __init__(self, dictApi, contentDefHelper=None, databaseName=None, **kwargs):
        """
        Args:
            dictApi (object): instance of DictionaryApi() class
            contentHelper (object, optional): an instance of a contentHelper().
            databaseName (string, optional): name of a content database (e.g. 'chem_comp', 'bird', 'bird_family', 'pdbx')

        """
        self.__dApi = dictApi
        self.__databaseName = databaseName
        self.__setup(contentDefHelper, databaseName)
        _ = kwargs
        #

    def __setup(self, contentDefHelper, databaseName):
        #
        iTypeCodes = []
        iQueryStrings = []
        unitCardinalityList = []
        dataSelectFilterD = OrderedDict()

        itemTransformD = OrderedDict()
        self.__categoryContentClasses = OrderedDict()
        self.__attributeContentClasses = OrderedDict()
        self.__intEnumD = OrderedDict()
        iterableD = OrderedDict()
        #
        self.__categoryList = sorted(self.__dApi.getCategoryList())
        self.__categorySchema = {catName: sorted(self.__dApi.getAttributeNameList(catName)) for catName in self.__categoryList}
        ##
        _, attributeContextIndex = self.__indexContexts(self.__categorySchema)
        self.__keyReplaceItems = attributeContextIndex["RCSB_CATEGORY_KEY_SUBSTITION"] if "RCSB_CATEGORY_KEY_SUBSTITION" in attributeContextIndex else []
        self.__keyReplaceCategoryD = OrderedDict()
        for catName, atName in self.__keyReplaceItems:
            self.__keyReplaceCategoryD.setdefault(catName, []).append(atName)
        #
        logger.debug("Primary key replacements: %r", self.__keyReplaceItems)
        logger.debug("Primary key replacement category index: %r", self.__keyReplaceCategoryD)
        #
        if contentDefHelper and databaseName:
            self.__intEnumD = {(tD["CATEGORY_NAME"], tD["ATTRIBUTE_NAME"]): True for tD in contentDefHelper.getInternalEnumItems(databaseName)}
            logger.debug("Internal enum items %r", self.__intEnumD)
            #
            cardD = contentDefHelper.getCardinalityKeyItem(databaseName)
            logger.debug("Cardinality attribute %r", cardD.items())
            #
            unitCardinalityList = self.__getUnitCardinalityCategories([cardD])
            unitCardinalityList.extend(contentDefHelper.getCardinalityCategoryExtras())
            logger.debug("Cardinality categories %r", unitCardinalityList)
            #
            self.__categoryContentClasses = contentDefHelper.getCategoryContentClasses(databaseName)
            logger.debug("categoryContentClasses %r", self.__categoryContentClasses)
            #
            self.__attributeContentClasses = contentDefHelper.getAttributeContentClasses(databaseName)
            logger.debug("attributeContentClasses %r", self.__attributeContentClasses)
            #
            self.__sliceParentItemsD = contentDefHelper.getDatabaseSliceParents(databaseName)
            self.__sliceUnitCardinalityD = OrderedDict()
            self.__sliceCategoryExtrasD = OrderedDict()
            for sliceName, pDL in self.__sliceParentItemsD.items():
                logger.debug("Slicename %s parents %r", sliceName, pDL)
                #
                # Some categories are include in a slice even if they are unconnected to the slice parent.
                self.__sliceCategoryExtrasD[sliceName] = contentDefHelper.getSliceCategoryExtras(databaseName, sliceName)
                logger.debug("Slice extra categories %r", self.__sliceCategoryExtrasD[sliceName])
                #
                self.__sliceUnitCardinalityD[sliceName] = self.__getUnitCardinalityCategories(pDL)
                logger.debug("Slice Unit cardinality categories %r", self.__sliceUnitCardinalityD[sliceName])
                #
                self.__sliceUnitCardinalityD[sliceName].extend(contentDefHelper.getSliceCardinalityCategoryExtras(databaseName, sliceName))
                logger.debug("Slicename %s unit cardinality categories %r", sliceName, self.__sliceUnitCardinalityD[sliceName])
            #
            #
            subCategoryD = {}
            for catName, atNameList in self.__categorySchema.items():
                scL = []
                for atName in atNameList:
                    scL.extend(self.__dApi.getItemSubCategoryIdList(catName, atName))
                scL = list(OrderedDict.fromkeys(scL)) if scL else []
                subCategoryD[catName] = scL
            logger.debug("Subcategory category dictionary %r", subCategoryD)
            #
            self.__categoryFeatures = {catName: self.__getCategoryFeatures(catName, unitCardinalityList, subCategoryD) for catName in self.__categoryList}
            iTypeCodes = contentDefHelper.getTypeCodes("iterable")
            emiTypeCodes = contentDefHelper.getTypeCodes("embedded_iterable")
            logger.debug("emiTypeCodes %r", emiTypeCodes)
            iQueryStrings = contentDefHelper.getQueryStrings("iterable")
            logger.debug("iterable types %r iterable query %r", iTypeCodes, iQueryStrings)
            iterableD = self.__getIterables(iTypeCodes, iQueryStrings, contentDefHelper)
            embeddedIterableD = self.__getEmbeddedIterables(emiTypeCodes)
            logger.debug("iterableD %r", iterableD.items())
            logger.debug("embeddedIterableD %r", embeddedIterableD.items())
            dataSelectFilterD = contentDefHelper.getDatabaseSelectionFilters(databaseName)
            itemTransformD = contentDefHelper.getItemTransformD()
            logger.debug("itemTransformD %r", itemTransformD.items())
            #
            self.__selectionFiltersD = contentDefHelper.getDatabaseSelectionFilters(databaseName)
            self.__sliceParentFiltersD = contentDefHelper.getDatabaseSliceParentFilters(databaseName)
        else:
            logger.debug("Dictionary helper not loaded for schema %r", databaseName, stack_info=True)
            self.__selectionFiltersD = {}
            self.__sliceParentItemsD = {}
            self.__sliceParentFiltersD = {}
            self.__sliceUnitCardinalityD = {}
            self.__sliceCategoryExtrasD = {}
            embeddedIterableD = {}
            logger.warning("Missing dictionary helper method or schema %r", databaseName)

        #
        self.__methodD = self.__getMethodInfo()
        self.__attributeFeatures = OrderedDict(
            {catName: self.__getAttributeFeatures(catName, iterableD, embeddedIterableD, itemTransformD, self.__methodD) for catName in self.__categoryList}
        )
        self.__attributeDataTypeD = OrderedDict({catName: self.__getAttributeTypeD(catName) for catName in self.__categoryList})
        self.__dataSelectFilterD = dataSelectFilterD
        self.__sliceD = self.__getSliceChildren(self.__sliceParentItemsD)

    def getSelectionFiltersForDatabase(self):
        try:
            return self.__selectionFiltersD
        except Exception:
            pass
        return {}

    def getSliceParentFiltersForDatabase(self):
        try:
            return self.__sliceParentFiltersD
        except Exception:
            pass
        return {}

    def getSliceParentItemsForDatabase(self):
        try:
            return self.__sliceParentItemsD
        except Exception:
            pass
        return {}

    def getSliceCategoryExtrasForDatabase(self):
        try:
            return self.__sliceCategoryExtrasD
        except Exception:
            pass
        return {}

    def getSliceUnitCardinalityForDatabase(self):
        try:
            return self.__sliceUnitCardinalityD
        except Exception:
            pass
        return {}

    def __getContentClasses(self, catName, atName=None, wildCardAtName="__all__"):
        """Return a list of contexts for input category and optional attribute.

        Handle the special case of unspecified attributes interpreted as wildcard.

        Return:
          contextList (list): list of context names
        """
        cL = []
        try:
            if atName is None:
                cL = self.__categoryContentClasses[catName] if catName in self.__categoryContentClasses else []
            elif (catName, wildCardAtName) in self.__attributeContentClasses:
                cL = self.__attributeContentClasses[(catName, wildCardAtName)]
            else:
                cL = self.__attributeContentClasses[(catName, atName)] if (catName, atName) in self.__attributeContentClasses else []
        except Exception as e:
            logger.exception("Failing catName %s atName %s with %s", catName, atName, str(e))
        #
        return cL

    def __getSliceChildren(self, sliceParentD):
        """Internal method to build data structure containing the parent-child relationships for the
        input slice parent construction.

        """
        retD = OrderedDict()
        for sliceName, sliceParents in sliceParentD.items():
            sD = OrderedDict()
            for pD in sliceParents:
                parentCategoryName = pD["CATEGORY_NAME"]
                parentAttributeName = pD["ATTRIBUTE_NAME"]
                #
                sD[parentCategoryName] = [{"PARENT_CATEGORY_NAME": parentCategoryName, "PARENT_ATTRIBUTE_NAME": parentAttributeName, "CHILD_ATTRIBUTE_NAME": parentAttributeName}]
                #
                # childItems = self.__dApi.getFullChildList(parentCategoryName, parentAttributeName)
                childItems = self.__dApi.getFullDescendentList(parentCategoryName, parentAttributeName)
                # logger.info("Slice parent %s %s  %r" % (parentCategoryName, parentAttributeName, childItems))
                for childItem in childItems:
                    atName = CifName.attributePart(childItem)
                    catName = CifName.categoryPart(childItem)
                    # Ignore children in the parent category
                    if catName == parentCategoryName:
                        continue
                    if catName not in sD:
                        sD[catName] = []
                    sD[catName].append({"PARENT_CATEGORY_NAME": parentCategoryName, "PARENT_ATTRIBUTE_NAME": parentAttributeName, "CHILD_ATTRIBUTE_NAME": atName})
                # Sort the list of dictionaries for each category
                for catName in sD:
                    sD[catName] = sorted(sD[catName], key=lambda k: (k["PARENT_CATEGORY_NAME"], k["PARENT_ATTRIBUTE_NAME"], k["CHILD_ATTRIBUTE_NAME"]))

            retD[sliceName] = sD
        return retD

    def __getIterables(self, iTypeCodes, iQueryStrings, dictHelper):
        itD = OrderedDict()
        #
        typD = {d["TYPE_CODE"]: d["DELIMITER"] for d in iTypeCodes}

        for catName in self.__categoryList:
            for atName in self.__categorySchema[catName]:
                typeCode = self.__dApi.getTypeCode(catName, atName)
                typeCodeAlt = self.__dApi.getTypeCodeAlt(catName, atName)
                if typeCode in typD or typeCodeAlt in typD:
                    itD[(catName, atName)] = typD[typeCode]
                    continue
                description = self.__dApi.getDescription(catName, atName)
                for qs in iQueryStrings:
                    if qs in description:
                        itD[(catName, atName)] = dictHelper.getDelimiter(catName, atName)
        logger.debug("iterableD: %r", list(itD.items()))
        return itD

    def __getEmbeddedIterables(self, iTypeCodes):
        itD = OrderedDict()
        #
        typD = {d["TYPE_CODE"]: d["DELIMITER"] for d in iTypeCodes}

        for catName in self.__categoryList:
            for atName in self.__categorySchema[catName]:
                typeCode = self.__dApi.getTypeCode(catName, atName)
                typeCodeAlt = self.__dApi.getTypeCodeAlt(catName, atName)
                if typeCode in typD or typeCodeAlt in typD:
                    itD[(catName, atName)] = typD[typeCode]
                    continue
        logger.debug("embedded iterableD: %r", list(itD.items()))
        return itD

    def __getMethodInfo(self):
        methodD = OrderedDict()
        methodIndex = self.__dApi.getMethodIndex()
        for _, mrL in methodIndex.items():
            for mr in mrL:
                mId = mr.getId()
                catName = mr.getCategoryName()
                atName = mr.getAttributeName()
                mType = mr.getType()
                logger.debug("mId %r catName %r atName %r mType %r", mId, catName, atName, mType)
                if (catName, atName) not in methodD:
                    methodD[(catName, atName)] = []
                methDef = self.__dApi.getMethod(mId)
                if methDef:
                    mLang = methDef.getLanguage()
                    mCode = methDef.getCode()
                    mImplement = methDef.getImplementation()
                    dD = {"METHOD_LANGUAGE": mLang, "METHOD_IMPLEMENT": mImplement, "METHOD_TYPE": mType, "METHOD_CODE": mCode}
                    methodD[(catName, atName)].append(dD)
                else:
                    logger.error("Missing method definition for %s", mId)
        ##
        logger.debug("Method dictionary %r", methodD)
        return methodD

    def getMethodImplementation(self, catName, atName, methodCodes=None):
        try:
            mC = methodCodes if methodCodes else ["calculation"]
            if (catName, atName) in self.__methodD:
                for mD in self.__methodD[(catName, atName)]:
                    if mD["METHOD_CODE"].lower() in mC:
                        return mD["METHOD_IMPLEMENT"]
            return None

        except Exception as e:
            logger.exception("Failing catName %s atName %s with %s", catName, atName, str(e))
        return None

    def getSelectionFilters(self):
        """
        Returns:  relevant data selection filters for in a dictionary by filter name.
        """
        return self.__dataSelectFilterD

    def getCategories(self):
        """
        Returns:  list of dictionary categories
        """
        #
        return self.__categoryList

    def getSchemaNames(self):
        return self.__categorySchema

    def getCategoryAttributes(self, catName):
        """

        Returns: list of attributes in the input category
        """
        return self.__categorySchema[catName]

    def getCategoryFeatures(self, catName):
        try:
            return self.__categoryFeatures[catName]
        except Exception as e:
            logger.error("Missing category %s %s", catName, str(e))
        return {}

    def getAttributeFeatures(self, catName):
        try:
            return self.__attributeFeatures[catName]
        except Exception as e:
            logger.error("Missing category %s %s", catName, str(e))
        return {}

    def getAttributeDataTypeD(self):
        return self.__attributeDataTypeD

    def getSliceNames(self):
        try:
            return list(self.__sliceD.keys())
        except Exception as e:
            logger.error("Failing with %s", str(e))
        return []

    def getSliceAttributes(self, sliceName, catName):
        try:
            return self.__sliceD[sliceName][catName]
        except Exception as e:
            logger.debug("Failing to access sliceName %s catName %s with %s", sliceName, catName, str(e))
        return []

    def getSliceCategories(self, sliceName):
        try:
            return list(self.__sliceD[sliceName].keys())
        except Exception as e:
            logger.debug("Failing to access sliceName %s with %s", sliceName, str(e))
        return []

    def __getAttributeTypeD(self, catName):
        aD = {}
        for atName in self.__categorySchema[catName]:
            aD[atName] = self.__dApi.getTypeCode(catName, atName)
        return aD

    def __assignEnumTypes(self, enumList, pT):
        rL = []
        if enumList:
            if pT != "numb":
                return enumList
            else:
                isFloat = False
                for enum in enumList:
                    if "." in enum:
                        isFloat = True
                        break
                if isFloat:
                    # a rare case
                    rL = [float(enum) for enum in enumList]
                else:
                    rL = [int(enum) for enum in enumList]
                return rL
        else:
            return enumList

    def __assignExampleTypes(self, catName, atName, examList, pT):

        try:
            rL = []
            if examList:
                if pT != "numb":
                    return examList
                else:
                    isFloat = False
                    for enum in examList:
                        if "." in enum:
                            isFloat = True
                            break
                    if isFloat:
                        # a rare case
                        rL = [float(enum) for enum in examList]
                    else:
                        rL = [int(enum) for enum in examList]
                    return rL

            else:
                return examList
        except Exception:
            logger.exception("catName %r atName %r primitive %r failing for examList %r", catName, atName, pT, examList)
        return examList

    def __assignExampleTupTypes(self, catName, atName, examTupList, pT):
        try:
            rL = []
            if examTupList:
                if pT != "numb":
                    for exam, detail in examTupList:
                        exam = str(exam).strip() if exam else exam
                        detail = str(detail).strip() if detail else detail
                        rL.append((exam, detail))
                    return rL
                else:
                    isFloat = False
                    for exam, _ in examTupList:
                        if "." in exam:
                            isFloat = True
                            break
                    if isFloat:
                        for exam, detail in examTupList:
                            try:
                                if isFloat:
                                    rL.append((float(exam), detail))
                                else:
                                    rL.append((int(exam), detail))
                            except Exception:
                                pass
                    return rL
            else:
                return examTupList
        except Exception as e:
            logger.exception("catName %r atName %r (%r) %r %s", catName, atName, pT, examTupList, str(e))
        #
        return examTupList

    def __assignEnumTupTypes(self, enumTupList, pT):
        rL = []
        if enumTupList:
            if pT != "numb":
                return enumTupList
            else:
                isFloat = False
                for enum, _, _, _ in enumTupList:
                    if "." in enum:
                        isFloat = True
                        break
                if isFloat:
                    for enum, detail, brief, units in enumTupList:
                        try:
                            if isFloat:
                                rL.append((float(enum), detail, brief, units))
                            else:
                                rL.append((int(enum), detail, brief, units))
                        except Exception:
                            pass
                return rL
        else:
            return enumTupList

    def __hasEnumDetails(self, enumTupList):
        dCount = 0
        for _, detail, brief, units in enumTupList:
            if detail or brief or units:
                dCount += 1
        return dCount > 0

    def __itemNameToDictList(self, itemNameList):
        rL = []
        for itemName in list(OrderedDict.fromkeys(itemNameList)):
            atName = CifName.attributePart(itemName)
            catName = CifName.categoryPart(itemName)
            rL.append({"CATEGORY": catName, "ATTRIBUTE": atName})
        return rL

    def __getAttributeFeatures(self, catName, iterableD, embeddedIterableD, itemTransformD, methodD):
        """
        Args:
            catName (string): Category name
            iterableD (tuple, optional): iterable dictionary type codes
            iQueryStrings (list, optional): search strings applied to item descriptions to identify iterable candidates
            itemTransformD (dict): dictionary of data transform filters  itd[(catName,atName)] = [f1,f2,...]

        Returns:
            dict: attribute features


             cL = self.getCategoryContextList(catName)
        """
        aD = {}

        #
        # keyAtNames = [CifName.attributePart(kyItem) for kyItem in self.__dApi.getCategoryKeyList(catName)]
        keyAtNames = [CifName.attributePart(kyItem) for kyItem in self.__getCategoryKeysWithReplacement(catName)]
        for atName in self.__categorySchema[catName]:
            itemName = CifName.itemName(catName, atName)
            fD = {
                "CATEGORY_NAME": catName,
                "ATTRIBUTE_NAME": atName,
                "TYPE_CODE": None,
                "TYPE_CODE_ALT": None,
                "IS_MANDATORY": False,
                "CHILD_ITEMS": [],
                "CHILDREN": [],
                "ROOT_PARENT_ITEM": None,
                "ROOT_PARENT": None,
                "PARENT": None,
                "DESCRIPTION": None,
                "DESCRIPTION_ANNOTATED": [],
                "IS_KEY": False,
                "ITERABLE_DELIMITER": None,
                "EMBEDDED_ITERABLE_DELIMITER": None,
                "FILTER_TYPES": [],
                "IS_CHAR_TYPE": False,
                "METHODS": [],
                "CONTENT_CLASSES": [],
                "UNITS": None,
                "ENUMS": None,
                "ENUMS_ANNOTATED": None,
                "SEARCH_CONTEXTS": None,
            }
            fD["TYPE_CODE"] = self.__dApi.getTypeCode(catName, atName)
            fD["TYPE_CODE_ALT"] = self.__dApi.getTypeCodeAlt(catName, atName)
            fD["IS_MANDATORY"] = True if str(self.__dApi.getMandatoryCode(catName, atName)).lower() in ["y", "yes"] else False
            fD["DESCRIPTION"] = textwrap.dedent(self.__dApi.getDescription(catName, atName)).lstrip().rstrip()
            #
            fD["DESCRIPTION_ANNOTATED"] = [{"text": fD["DESCRIPTION"], "context": "dictionary"}]
            tS = self.__dApi.getDescriptionPdbx(catName, atName)
            if tS:
                fD["DESCRIPTION_ANNOTATED"].append({"text": textwrap.dedent(tS).lstrip().rstrip(), "context": "deposition"})
            #
            fD["UNITS"] = self.__dApi.getUnits(catName, atName)

            #
            fD["CHILD_ITEMS"] = self.__dApi.getFullChildList(catName, atName)
            fD["CHILDREN"] = self.__itemNameToDictList(self.__dApi.getFullChildList(catName, atName))
            #
            pItemName = self.__dApi.getUltimateParent(catName, atName)
            pName = pItemName if pItemName != itemName else None
            fD["ROOT_PARENT_ITEM"] = pName

            fD["ROOT_PARENT"] = self.__itemNameToDictList([pName])[0] if pName else None
            #
            pL = self.__dApi.getFullParentList(catName, atName, stripSelfParent=True)
            if pL:
                rL = self.__itemNameToDictList(pL)
                fD["PARENT"] = rL[0] if rL else None
                if len(rL) > 1:
                    logger.warning("Unexpected multiple parent definition for %s %s : %r", catName, atName, rL)
            #
            # logger.debug("catName %s atName %s : parent %r root_parent %r", catName, atName, fD['PARENT'], fD['ROOT_PARENT'])
            #
            fD["IS_KEY"] = atName in keyAtNames
            pType = self.__dApi.getTypePrimitive(catName, atName)
            fD["IS_CHAR_TYPE"] = str(pType).lower() in ["char", "uchar"]
            #
            fD["ITERABLE_DELIMITER"] = iterableD[(catName, atName)] if (catName, atName) in iterableD else None
            fD["EMBEDDED_ITERABLE_DELIMITER"] = embeddedIterableD[(catName, atName)] if (catName, atName) in embeddedIterableD else None
            #
            fD["FILTER_TYPES"] = itemTransformD[(catName, "__all__")] if (catName, "__all__") in itemTransformD else []
            fD["FILTER_TYPES"] = itemTransformD[(catName, atName)] if (catName, atName) in itemTransformD else fD["FILTER_TYPES"]
            #
            fD["METHODS"] = methodD[(catName, atName)] if (catName, atName) in methodD else []
            fD["CONTENT_CLASSES"] = self.__getContentClasses(catName, atName)
            if (catName, atName) in self.__intEnumD:
                fD["ENUMS"] = sorted(self.__assignEnumTypes(self.__dApi.getEnumListPdbx(catName, atName), pType))
                logger.debug("Using internal enums for %s %s %d", catName, atName, len(fD["ENUMS"]))
                enumTupList = self.__dApi.getEnumListAltWithFullDetails(catName, atName)
            else:
                fD["ENUMS"] = sorted(self.__assignEnumTypes(self.__dApi.getEnumList(catName, atName), pType))
                enumTupList = self.__dApi.getEnumListWithFullDetails(catName, atName)
            #
            if self.__hasEnumDetails(enumTupList):
                #
                fD["ENUMS_ANNOTATED"] = []
                for eTup in self.__assignEnumTupTypes(enumTupList, pType):
                    teD = {"value": eTup[0]}
                    if eTup[1]:
                        teD["detail"] = eTup[1]
                    if eTup[2]:
                        teD["name"] = eTup[2]
                    if eTup[3]:
                        teD["units"] = eTup[3]
                    fD["ENUMS_ANNOTATED"].append(teD)
            # -----
            fD["EXAMPLES"] = self.__assignExampleTupTypes(catName, atName, self.__dApi.getExampleListPdbx(catName, atName), pType)
            fD["EXAMPLES"].extend(self.__assignExampleTupTypes(catName, atName, self.__dApi.getExampleList(catName, atName), pType))
            # -----
            scL = []
            for scTup in self.__dApi.getItemSubCategoryList(catName, atName):
                if scTup[1] is not None:
                    qD = {"id": scTup[0], "label": scTup[1]}
                else:
                    qD = {"id": scTup[0]}
                scL.append(qD)
            fD["SUB_CATEGORIES"] = scL
            if len(scL) > 1:
                logger.debug("Multiple subcategories for %r %r %r", catName, atName, scL)
            #
            # bList = self.__dApi.getBoundaryListAlt(catName, atName, fallBack=True)
            bdList = self.__dApi.getBoundaryList(catName, atName)
            if bdList:
                minD = {}
                maxD = {}
                for (minV, maxV) in bdList:
                    if minV == maxV:
                        continue
                    if minV not in [".", "?"]:
                        minD[minV] = False
                    if maxV not in [".", "?"]:
                        maxD[maxV] = False
                for (minV, maxV) in bdList:
                    if minV == maxV and minV in minD:
                        minD[minV] = True
                    if minV == maxV and maxV in maxD:
                        maxD[maxV] = True
                for ky in minD:
                    if "." in ky:
                        kyV = float(ky)
                    else:
                        kyV = int(ky)
                    if minD[ky]:
                        fD["MIN_VALUE"] = kyV
                    else:
                        fD["MIN_VALUE_EXCLUSIVE"] = kyV
                for ky in maxD:
                    if "." in ky:
                        kyV = float(ky)
                    else:
                        kyV = int(ky)
                    if maxD[ky]:
                        fD["MAX_VALUE"] = kyV
                    else:
                        fD["MAX_VALUE_EXCLUSIVE"] = kyV
            #
            aD[atName] = fD
        #
        return aD

    def __getCategoryFeatures(self, catName, unitCardinalityList, subCategoryD):
        cD = {"KEY_ATTRIBUTES": []}
        # cD['KEY_ATTRIBUTES'] = [CifName.attributePart(keyItem) for keyItem in self.__dApi.getCategoryKeyList(catName)]
        cD["KEY_ATTRIBUTES"] = [CifName.attributePart(keyItem) for keyItem in self.__getCategoryKeysWithReplacement(catName)]
        cD["UNIT_CARDINALITY"] = catName in unitCardinalityList
        cD["CONTENT_CLASSES"] = self.__getContentClasses(catName)
        cD["IS_MANDATORY"] = True if str(self.__dApi.getCategoryMandatoryCode(catName)).lower() == "yes" else False
        cD["SUB_CATEGORIES"] = subCategoryD[catName] if catName in subCategoryD else []
        #
        return cD

    def __getUnitCardinalityCategories(self, parentDList):
        """Assign categories with unit cardinality relative to the input list of parent key items.

        parentDList (dict):  [{'CATEGORY_NAME':xxx 'ATTRIBUTE_NAME': xxxx}]

        Return: category name list
        """
        numParents = len(parentDList)
        logger.debug("Parent slice count %d def %r", numParents, parentDList)
        ucL = []
        #
        #  Find the common set of child categories for the input parent items
        comCatList = []
        for pD in parentDList:
            catList = [pD["CATEGORY_NAME"]]
            for childItem in self.__dApi.getFullChildList(pD["CATEGORY_NAME"], pD["ATTRIBUTE_NAME"]):
                childCategoryName = CifName.categoryPart(childItem)
                primaryKeyItemList = self.__dApi.getCategoryKeyList(childCategoryName)
                logger.debug("child category %r primary key items  %r", childCategoryName, primaryKeyItemList)
                # child must be part of the primary key to be a candidate
                if childItem in primaryKeyItemList:
                    catList.append(childCategoryName)
            if comCatList:
                comCatList = list(set(catList) & set(comCatList))
            else:
                comCatList.extend(catList)
        logger.debug("Common category list %r", comCatList)
        for cat in comCatList:
            primaryKeyItemList = self.__dApi.getCategoryKeyList(cat)
            if len(primaryKeyItemList) == numParents:
                ucL.append(cat)
        #
        logger.debug("Slice unit cardinality categories from parent-child relationships %r", ucL)
        return sorted(ucL)

    #

    def __getCategoryKeysWithReplacement(self, categoryName):
        if categoryName in self.__keyReplaceCategoryD:
            keyItems = [CifName.itemName(categoryName, atName) for atName in self.__keyReplaceCategoryD[categoryName]]
        else:
            keyItems = self.__dApi.getCategoryKeyList(categoryName)
        return sorted(keyItems)

    def __indexContexts(self, dictSchema):
        """Extract the category an item level dictionary contexts."""
        catIndex = {}
        atIndex = {}
        for catName in dictSchema:
            for cT in self.__dApi.getCategoryContextList(catName):
                catIndex.setdefault(cT, []).append(catName)

            for atName in dictSchema[catName]:
                for cT in self.__dApi.getContextList(catName, atName):
                    atIndex.setdefault(cT, []).append((catName, atName))
        return catIndex, atIndex



##
# File: SequenceClustersEtlWorker.py
# Date: 2-Jul-2018  jdw
#
# ETL utilities for sequence cluster data
#
# Updates:
#  15-Jul-2018 jdw split out to separate module and add status tracking
#  28-Oct-2018 jdw adjustments for new configuration organization
#   4-Jan-2019 jdw differentiate site and application config sections for provenance.
#
##
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging

from rcsb.db.mongo.DocumentLoader import DocumentLoader
from rcsb.db.processors.ClusterDataPrep import ClusterDataPrep
from rcsb.db.processors.DataExchangeStatus import DataExchangeStatus
from rcsb.db.utils.ProvenanceProvider import ProvenanceProvider

logger = logging.getLogger(__name__)


class SequenceClustersEtlWorker(object):
    """Prepare and load sequence cluster data.

     Note: relevant configuration options -

    site_info:
     'RCSB_SEQUENCE_CLUSTER_DATA_PATH': ...

    entity_sequence_clusters:
     DATABASE_NAME: sequence_clusters
     DATABASE_VERSION_STRING: v5
     COLLECTION_ENTITY_MEMBERS: entity_members
     COLLECTION_ENTITY_MEMBERS_INDEX: data_set_id,entry_id,entity_id
     COLLECTION_CLUSTER_MEMBERS: cluster_members
     COLLECTION_CLUSTER_MEMBERS_INDEX: data_set_id,identity,cluster_id
     COLLECTION_VERSION_STRING: v0_1
     ENTITY_SCHEMA_NAME: rcsb_entity_sequence_cluster_entity_list
     CLUSTER_SCHEMA_NAME: rcsb_entity_sequence_cluster_identifer_list
     SEQUENCE_IDENTITY_LEVELS: 100,95,90,70,50,30
     COLLECTION_CLUSTER_PROVENANCE: cluster_provenance
     PROVENANCE_KEY_NAME: rcsb_entity_sequence_cluster_prov
     PROVENANCE_INFO_LOCATOR: provenance/rcsb_extend_provenance_info.json


    """

    def __init__(self, cfgOb, workPath=None, numProc=2, chunkSize=10, readBackCheck=False, documentLimit=None, verbose=False):
        self.__cfgOb = cfgOb
        self.__cachePath = workPath
        self.__readBackCheck = readBackCheck
        self.__numProc = numProc
        self.__chunkSize = chunkSize
        self.__documentLimit = documentLimit
        self.__resourceName = "MONGO_DB"
        self.__verbose = verbose
        #
        self.__sectionCluster = "entity_sequence_clusters_configuration"
        self.__clusterDataPath = self.__cfgOb.getPath("RCSB_SEQUENCE_CLUSTER_DATA_PATH", sectionName=self.__cfgOb.getDefaultSectionName())
        self.__databaseName = self.__cfgOb.get("DATABASE_NAME", sectionName=self.__sectionCluster, default="sequence_clusters")
        self.__databaseVersion = self.__cfgOb.get("DATABASE_VERSION_STRING", sectionName=self.__sectionCluster, default="v5")
        #
        self.__entityMemberCollection = self.__cfgOb.get("COLLECTION_ENTITY_MEMBERS", sectionName=self.__sectionCluster, default="entity_members")
        self.__clusterMembersCollection = self.__cfgOb.get("COLLECTION_CLUSTER_MEMBERS", sectionName=self.__sectionCluster, default="cluster_members")
        self.__clusterProvenanceCollection = self.__cfgOb.get("COLLECTION_CLUSTER_PROVENANCE", sectionName=self.__sectionCluster, default="cluster_provenance")
        # self.__collectionVersion = self.__cfgOb.get("COLLECTION_VERSION_STRING", sectionName=self.__sectionCluster, default="v0_1")
        #
        self.__entitySchemaName = self.__cfgOb.get("ENTITY_SCHEMA_NAME", sectionName=self.__sectionCluster, default="rcsb_entity_sequence_cluster_entity_list")
        self.__clusterSchemaName = self.__cfgOb.get("CLUSTER_SCHEMA_NAME", sectionName=self.__sectionCluster, default="rcsb_entity_sequence_cluster_identifer_list")
        #
        tS = self.__cfgOb.get("COLLECTION_ENTITY_MEMBERS_INDEX", sectionName=self.__sectionCluster, default=None)
        self.__entityMemberCollectionIndexL = tS.split(",") if tS else None
        tS = self.__cfgOb.get("COLLECTION_CLUSTER_MEMBERS_INDEX", sectionName=self.__sectionCluster, default=None)
        self.__clusterMembersCollectionIndexL = tS.split(",") if tS else None
        # sample data set
        #
        tS = self.__cfgOb.get("SEQUENCE_IDENTITY_LEVELS", sectionName=self.__sectionCluster, default=None)
        self.__identityLevels = tS.split(",") if tS else ["100", "95", "90", "70", "50", "30"]
        #
        self.__statusList = []
        #

    def __updateStatus(self, updateId, databaseName, collectionName, status, startTimestamp):
        try:
            sFlag = "Y" if status else "N"
            desp = DataExchangeStatus()
            desp.setStartTime(tS=startTimestamp)
            desp.setObject(databaseName, collectionName)
            desp.setStatus(updateId=updateId, successFlag=sFlag)
            desp.setEndTime()
            self.__statusList.append(desp.getStatus())
            return True
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return False

    def __extract(self, dataSetId, dataLocator, levels):
        """Extract sequence cluster data set  (mmseq2 or blastclust organization)"""
        try:
            cdp = ClusterDataPrep(workPath=self.__cachePath, entitySchemaName=self.__entitySchemaName, clusterSchemaName=self.__clusterSchemaName)
            _, docBySequenceD, docByClusterD = cdp.extract(dataSetId, clusterSetLocator=dataLocator, levels=levels, clusterType="entity")
            return docBySequenceD, docByClusterD
        except Exception as e:
            logger.exception("Failing with %s", str(e))

        return {}, {}

    def __fetchProvenance(self):
        """Fetching a provenance dictionary content."""
        try:
            provKeyName = self.__cfgOb.get("PROVENANCE_KEY_NAME", sectionName=self.__sectionCluster, default="rcsb_entity_sequence_cluster_prov")
            provU = ProvenanceProvider(self.__cfgOb, self.__cachePath, useCache=True)
            pD = provU.fetch()
            return pD[provKeyName] if provKeyName in pD else {}
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return {}

    def etl(self, dataSetId, dataLocator=None, loadType="full"):
        """Prepare and load sequence cluster data by entity and by cluster identifer."""
        try:
            self.__statusList = []
            desp = DataExchangeStatus()
            statusStartTimestamp = desp.setStartTime()
            #
            docBySequenceD, docByClusterD = self.__extract(dataSetId=dataSetId, dataLocator=dataLocator, levels=self.__identityLevels)
            #
            dl = DocumentLoader(
                self.__cfgOb,
                self.__cachePath,
                self.__resourceName,
                numProc=self.__numProc,
                chunkSize=self.__chunkSize,
                documentLimit=self.__documentLimit,
                verbose=self.__verbose,
                readBackCheck=self.__readBackCheck,
            )
            #
            databaseName = self.__databaseName
            # addValues = {"_schema_version": self.__collectionVersion}
            addValues = None
            #
            collectionName = self.__entityMemberCollection
            dList = docBySequenceD[self.__entitySchemaName]
            ok1 = dl.load(
                databaseName, collectionName, loadType=loadType, documentList=dList, indexAttributeList=self.__entityMemberCollectionIndexL, keyNames=None, addValues=addValues
            )
            self.__updateStatus(dataSetId, databaseName, collectionName, ok1, statusStartTimestamp)
            #
            collectionName = self.__clusterMembersCollection
            dList = docByClusterD[self.__clusterSchemaName]
            ok2 = dl.load(
                databaseName, collectionName, loadType=loadType, documentList=dList, indexAttributeList=self.__clusterMembersCollectionIndexL, keyNames=None, addValues=addValues
            )
            self.__updateStatus(dataSetId, databaseName, collectionName, ok2, statusStartTimestamp)
            #
            pD = self.__fetchProvenance()
            collectionName = self.__clusterProvenanceCollection
            ok3 = dl.load(databaseName, collectionName, loadType=loadType, documentList=[pD], indexAttributeList=None, keyNames=None, addValues=addValues)
            self.__updateStatus(dataSetId, databaseName, collectionName, ok3, statusStartTimestamp)
            #
            return ok1 and ok2 and ok3
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return False

    def getLoadStatus(self):
        return self.__statusList



#
# from six.moves.html_parser import HTMLParser

try:
    # Python 2.6-2.7
    from HTMLParser import unescape
except ImportError:
    # Python 3.6+
    from html import unescape


def unescapeXmlCharRef(iStr):
    """
    Convert html character entities into unicode.
    """
    oStr = unescape(iStr)

    return oStr


if __name__ == "__main__":
    print("START")
    print("%r" % unescapeXmlCharRef("&lt;b&gt;"))
    print("%r" % unescapeXmlCharRef("Here is a &quot;").encode("utf-8"))
    print("%r" % unescapeXmlCharRef("Here is a &Phi;").encode("utf-8"))
    print("%range" % unescapeXmlCharRef("Here is a &Psi;"))
    print("%r" % unescapeXmlCharRef("Here is a &alpha;"))
    print("%r" % unescapeXmlCharRef("Here is a &#xa3;"))

    print("%r" % unescapeXmlCharRef("Here is a &#8453;"))
    print("%r" % unescapeXmlCharRef("Here is a &#9734;"))
    print("%r" % unescapeXmlCharRef("Here is a &#120171;"))
    for ichar in range(1, 8000):
        myStr = "decimal %6d char &#%d;" % (ichar, ichar)
        print("%r" % unescapeXmlCharRef(myStr))



##
# File: RepoLoadWorkflow.py
# Date: 15-Dec-2019  jdw
#
#  Workflow wrapper  --  repository database loading utilities --
#
#  Updates:
#
##
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging
import os

from rcsb.db.cli.RepoHoldingsEtlWorker import RepoHoldingsEtlWorker
from rcsb.db.cli.SequenceClustersEtlWorker import SequenceClustersEtlWorker
from rcsb.utils.dictionary.DictMethodResourceProvider import DictMethodResourceProvider
from rcsb.db.mongo.DocumentLoader import DocumentLoader
from rcsb.db.mongo.PdbxLoader import PdbxLoader
from rcsb.db.utils.TimeUtil import TimeUtil
from rcsb.utils.config.ConfigUtil import ConfigUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil

logger = logging.getLogger(__name__)


class RepoLoadWorkflow(object):
    def __init__(self, **kwargs):
        #  Configuration Details
        configPath = kwargs.get("configPath", "exdb-config-example.yml")
        self.__configName = kwargs.get("configName", "site_info_configuration")
        mockTopPath = kwargs.get("mockTopPath", None)
        self.__cfgOb = ConfigUtil(configPath=configPath, defaultSectionName=self.__configName, mockTopPath=mockTopPath)
        #
        self.__cachePath = kwargs.get("cachePath", ".")
        self.__cachePath = os.path.abspath(self.__cachePath)
        self.__debugFlag = kwargs.get("debugFlag", False)
        if self.__debugFlag:
            logger.setLevel(logging.DEBUG)
        #
        #  Rebuild or check resource cache
        # rebuildCache = kwargs.get("rebuildCache", False)
        # self.__cacheStatus = self.buildResourceCache(rebuildCache=rebuildCache)
        # logger.debug("Cache status if %r", self.__cacheStatus)
        #

    def load(self, op, **kwargs):
        # if not self.__cacheStatus:
        #    logger.error("Resource cache test or rebuild has failed - exiting")
        #    return False
        # argument processing
        if op not in ["pdbx-loader", "etl-repository-holdings", "etl-entity-sequence-clusters"]:
            logger.error("Unsupported operation %r - exiting", op)
            return False
        try:
            readBackCheck = kwargs.get("readBackCheck", False)
            numProc = int(kwargs.get("numProc", 1))
            chunkSize = int(kwargs.get("chunkSize", 10))
            fileLimit = int(kwargs.get("fileLimit")) if "fileLimit" in kwargs else None
            documentLimit = int(kwargs.get("documentLimit")) if "documentLimit" in kwargs else None
            failedFilePath = kwargs.get("failFileListPath", None)
            loadFileListPath = kwargs.get("loadFileListPath", None)
            saveInputFileListPath = kwargs.get("saveFileListPath", None)
            schemaLevel = kwargs.get("schemaLevel", "min") if kwargs.get("schemaLevel") in ["min", "full"] else "min"
            loadType = kwargs.get("loadType", "full")  # or replace
            updateSchemaOnReplace = kwargs.get("updateSchemaOnReplace", True)
            pruneDocumentSize = float(kwargs.get("pruneDocumentSize")) if "pruneDocumentSize" in kwargs else None

            # "Document organization (rowwise_by_name_with_cardinality|rowwise_by_name|columnwise_by_name|rowwise_by_id|rowwise_no_name",
            documentStyle = kwargs.get("documentStyle", "rowwise_by_name_with_cardinality")
            dbType = kwargs.get("dbType", "mongo")
            #
            databaseName = kwargs.get("databaseName", None)
            databaseNameList = self.__cfgOb.get("DATABASE_NAMES_ALL", sectionName="database_catalog_configuration").split(",")
            collectionNameList = kwargs.get("collectionNameList", None)
            mergeValidationReports = kwargs.get("mergeValidationReports", True)
            #
            tU = TimeUtil()
            dataSetId = kwargs.get("dataSetId") if "dataSetId" in kwargs else tU.getCurrentWeekSignature()
            seqDataLocator = self.__cfgOb.getPath("RCSB_SEQUENCE_CLUSTER_DATA_PATH", sectionName=self.__configName)
            sandboxPath = self.__cfgOb.getPath("RCSB_EXCHANGE_SANDBOX_PATH", sectionName=self.__configName)

        except Exception as e:
            logger.exception("Argument and configuration processing failing with %s", str(e))
            return False
        #

        if op == "pdbx-loader" and dbType == "mongo" and databaseName in databaseNameList:
            okS = True
            try:
                inputPathList = None
                if loadFileListPath:
                    mu = MarshalUtil(workPath=self.__cachePath)
                    inputPathList = mu.doImport(loadFileListPath, fmt="list")
                    if not inputPathList:
                        logger.error("Operation %r missing or empty input file path list %s - exiting", op, loadFileListPath)
                        return False
            except Exception as e:
                logger.exception("Operation %r processing input path list failing with %s", op, str(e))
                return False
            #
            try:
                mw = PdbxLoader(
                    self.__cfgOb,
                    self.__cachePath,
                    resourceName="MONGO_DB",
                    numProc=numProc,
                    chunkSize=chunkSize,
                    fileLimit=fileLimit,
                    verbose=self.__debugFlag,
                    readBackCheck=readBackCheck,
                )
                ok = mw.load(
                    databaseName,
                    collectionLoadList=collectionNameList,
                    loadType=loadType,
                    inputPathList=inputPathList,
                    styleType=documentStyle,
                    dataSelectors=["PUBLIC_RELEASE"],
                    failedFilePath=failedFilePath,
                    saveInputFileListPath=saveInputFileListPath,
                    pruneDocumentSize=pruneDocumentSize,
                    validationLevel=schemaLevel,
                    mergeContentTypes=["vrpt"] if mergeValidationReports else None,
                    updateSchemaOnReplace=updateSchemaOnReplace,
                )
                okS = self.loadStatus(mw.getLoadStatus(), readBackCheck=readBackCheck)
            except Exception as e:
                logger.exception("Operation %r database %r failing with %s", op, databaseName, str(e))
        elif op == "etl-entity-sequence-clusters" and dbType == "mongo":
            cw = SequenceClustersEtlWorker(
                self.__cfgOb, numProc=numProc, chunkSize=chunkSize, documentLimit=documentLimit, verbose=self.__debugFlag, readBackCheck=readBackCheck, workPath=self.__cachePath
            )
            ok = cw.etl(dataSetId, seqDataLocator, loadType=loadType)
            okS = self.loadStatus(cw.getLoadStatus(), readBackCheck=readBackCheck)
        elif op == "etl-repository-holdings" and dbType == "mongo":
            rhw = RepoHoldingsEtlWorker(
                self.__cfgOb,
                sandboxPath,
                self.__cachePath,
                numProc=numProc,
                chunkSize=chunkSize,
                documentLimit=documentLimit,
                verbose=self.__debugFlag,
                readBackCheck=readBackCheck,
            )
            ok = rhw.load(dataSetId, loadType=loadType)
            okS = self.loadStatus(rhw.getLoadStatus(), readBackCheck=readBackCheck)

        logger.info("Completed operation %r with status %r", op, ok and okS)

        return ok and okS

    def loadStatus(self, statusList, readBackCheck=True):
        ret = False
        try:
            dl = DocumentLoader(self.__cfgOb, self.__cachePath, "MONGO_DB", numProc=1, chunkSize=2, documentLimit=None, verbose=False, readBackCheck=readBackCheck)
            #
            sectionName = "data_exchange_configuration"
            databaseName = self.__cfgOb.get("DATABASE_NAME", sectionName=sectionName)
            collectionName = self.__cfgOb.get("COLLECTION_UPDATE_STATUS", sectionName=sectionName)
            ret = dl.load(databaseName, collectionName, loadType="append", documentList=statusList, indexAttributeList=["update_id", "database_name", "object_name"], keyNames=None)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return ret

    def buildResourceCache(self, rebuildCache=False):
        """Generate and cache resource dependencies."""
        ret = False
        try:
            useCache = not rebuildCache
            rP = DictMethodResourceProvider(
                self.__cfgOb,
                configName=self.__configName,
                cachePath=self.__cachePath,
                restoreUseStash=True,
                restoreUseGit=True,
                providerTypeExclude=None,
            )
            ret = rP.cacheResources(useCache=useCache, doBackup=False, useStash=False, useGit=False)
            logger.info("useCache %r cache reload status (%r)", useCache, ret)

        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return ret



##
# File: MysqlSchemaImporter.py
# Date: 20-May-2015 jdw
#
# Create a skeleton schema map definition data structure from mysql
# table descriptions.
#
# ** This is a command-line utility that is not part of the overall api
#
import logging
import os
import pprint
import sys

logging.basicConfig(level=logging.INFO, format="%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s")
logger = logging.getLogger()
logger.setLevel(logging.INFO)
logger = logging.getLogger(__name__)


class MysqlSchemaImporter(object):
    def __init__(self, dbUser, dbPw, dbHost, mysqlPath="/opt/local/bin/mysql", verbose=True):
        self.__verbose = verbose
        self.__mysqlPath = mysqlPath
        self.__dbUser = dbUser
        self.__dbPw = dbPw
        self.__dbHost = dbHost

    def __import(self, filePath):
        colDataList = []
        ifh = open(filePath, "r")
        for line in ifh:
            if line is not None and line:
                fields = str(line[:-1]).split("\t")
                if len(fields) != 6:
                    logger.info("bad line in %s  = %s", filePath, line)
                    continue
                else:
                    colDataList.append(fields)
        ifh.close()
        os.remove(filePath)
        #
        return colDataList[1:]

    def __export(self, filePath, db, tableName):
        cmdDetail = ' --user=%s --password=%s --host=%s %s -e "describe %s;" ' % (self.__dbUser, self.__dbPw, self.__dbHost, db, tableName)
        cmd = self.__mysqlPath + cmdDetail + " > %s" % filePath
        return os.system(cmd)

    def __buildDef(self, dbName, tableName, colDataList):
        _ = dbName
        defD = {}
        tableId = str(tableName).upper()
        attIdKeyList = []
        attMap = {}
        indD = {}
        attInfo = {}
        attD = {}
        for ii, ff in enumerate(colDataList, start=1):
            attName = str(ff[0])
            attId = str(attName).upper()
            nullFlag = True if ff[2] == "YES" else False
            impType = ff[1]
            if "(" in impType:
                width = impType[impType.find("(") + 1 : -1]
                sqlType = impType[: impType.find("(")]
            else:
                width = 10
                sqlType = impType
            precision = 0
            if ff[3] in ["MUL", "PRI"]:
                attIdKeyList.append(attId)
                keyFlag = True
            else:
                keyFlag = False
            attD[attId] = attName
            attMap[attId] = (tableName, attName, None, None)
            attInfo[attId] = {"NULLABLE": nullFlag, "ORDER": ii, "PRECISION": precision, "PRIMARY_KEY": keyFlag, "APP_TYPE": sqlType.upper(), "WIDTH": width}
        #
        dD = {}
        dD["ATTRIBUTES"] = tuple(attIdKeyList)
        dD["TYPE"] = "UNIQUE"
        indD["p1"] = dD
        defD["INDICES"] = indD
        defD["ATTRIBUTES"] = attD
        defD["ATTRIBUTE_INFO"] = attInfo
        defD["ATTRIBUTE_MAP"] = attMap
        #
        defD["SCHEMA_DELETE_ATTRIBUTE"] = attIdKeyList[0]
        defD["SCHEMA_ID"] = tableId
        defD["SCHEMA_NAME"] = tableName
        defD["SCHEMA_TYPE"] = "transactional"
        # 'MAP_MERGE_INDICES': {'valence_ref': {'ATTRIBUTES': ('id',), 'TYPE': 'EQUI-JOIN'}},
        tD = {}
        tD["ATTRIBUTES"] = tuple(attIdKeyList)
        tD["TYPE"] = "EQUI-JOIN"
        defD["MAP_MERGE_INDICES"] = {}
        defD["MAP_MERGE_INDICES"][tableName] = tD
        #
        return tableId, defD

    def create(self, dbName, tableNameList):
        schemaDef = {}
        for tableName in tableNameList:
            fn = "mysql-schema-" + tableName + ".txt"
            self.__export(fn, dbName, tableName)
            colDataList = self.__import(fn)
            if colDataList:
                logger.info("tableName %s length %d\n", tableName, len(colDataList))
                tableId, defD = self.__buildDef(dbName, tableName, colDataList)
                schemaDef[tableId] = defD
        #
        pprint.pprint(schemaDef, stream=sys.stdout, width=120, indent=3)


def importExample():
    # tableNameList = ["entity", "entity_poly", "entity_src_gen", "entity_src_nat", "pdbx_entity_src_syn"]
    tableNameList = [
        "PDB_status_information",
        "audit_author",
        "chem_comp",
        "citation",
        "citation_author",
        "deposition_from_09",
        "deposition_from_2",
        "entity",
        "entity_poly",
        "pdb_entry",
        "pdbx_contact_author",
        "pdbx_database_PDB_obs_spr",
        "pdbx_database_status_history",
        "pdbx_depui_entry_details",
        "pdbx_molecule",
        "pdbx_molecule_features",
        "pdbx_prerelease_seq",
        "processing_status",
        "rcsb_status",
        "rcsb_status_t",
        "struct",
        "pdbx_database_status_history",
    ]
    # tableNameList = ['weight_in_asu']
    dbUser = os.getenv("MYSQL_DB_USER")
    dbPw = os.getenv("MYSQL_SBKB_PW")
    dbHost = "localhost"
    msi = MysqlSchemaImporter(dbUser, dbPw, dbHost, mysqlPath="/opt/local/bin/mysql", verbose=True)
    msi.create("stat", tableNameList)


if __name__ == "__main__":
    importExample()



##
# File:    SchemaDefLoader.py
# Author:  J. Westbrook
# Date:    7-Jan-2013
# Version: 0.001 Initial version
#
# Updates:
#  9-Jan-2013 jdw add merging index support for loading tables from multiple
#                 instance categories.
# 10-Jan-2013 jdw add null value filter and maximum string width checks.
# 13-Jan-2013 jdw provide batch file and batch insert loading modes.
# 15-Jan-2013 jdw add pre-load delete options
# 19-Jan-2012 jdw add IoAdapter
# 20-Jan-2013 jdw add append options for batch file loading
# 20-Jan-2013 jdw provide methods for loading container lists
#  2-Oct-2017 jdw escape null string '\N' and suppress print statements
# 20-Dec-2017 jdw set to use python adapter -
# 30-Dec-2017 jdw add crate server support - 'crate-insert', 'crate-insert-many'
#  4-Jan-2018 jdw add table skipping filters
#  4-Feb-2018 jdw add cockroach server support - 'cockroach-insert', 'cockroach-insert-many'
# 13-Mar-2018 jdw split data loading and data processing operations.
# 30-Mar-2018 jdw more refactoring -  changing expectations on explicit data types -
#  2-Jul-2018 jdw fix working path
# 20-Aug-2019 jdw add dynamic method invocation  -
# 11-Nov-2018 jdw  add DrugBank and CCDC mapping path details.
# 31-Mar-2019 jdw add more speific tests for null value suggested by
#                 issue = MySQL SchemaDefLoader skip zero values #19
##
"""
Generic mapper of PDBx/mmCIF instance data to SQL loadable data files based on external
schema definition defined in class SchemaDefBase().

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"


import csv
import logging
import os
import time

from mmcif.api.DictMethodRunner import DictMethodRunner
from rcsb.utils.dictionary.DictionaryApiProviderWrapper import DictionaryApiProviderWrapper
from rcsb.utils.dictionary.DictMethodResourceProvider import DictMethodResourceProvider
from rcsb.db.mysql.MyDbUtil import MyDbQuery
from rcsb.db.processors.DataTransformFactory import DataTransformFactory
from rcsb.db.processors.SchemaDefDataPrep import SchemaDefDataPrep
from rcsb.db.sql.SqlGen import SqlGenAdmin
from rcsb.utils.repository.RepositoryProvider import RepositoryProvider

logger = logging.getLogger(__name__)
#


class SchemaDefLoader(object):

    """Map PDBx/mmCIF instance data to SQL loadable data using external schema definition."""

    def __init__(
        self,
        cfgOb,
        schemaDefObj,
        cfgSectionName="site_info_configuration",
        dbCon=None,
        cachePath=".",
        workPath=".",
        cleanUp=False,
        warnings="default",
        verbose=True,
        restoreUseStash=True,
        restoreUseGit=True,
        providerTypeExclude=True,
    ):
        self.__verbose = verbose
        self.__debug = False
        self.__cfgOb = cfgOb
        sectionName = cfgSectionName
        self.__sD = schemaDefObj

        #
        self.__dbCon = dbCon
        self.__cachePath = cachePath
        self.__workPath = workPath
        self.__pathList = []
        self.__cleanUp = cleanUp
        #
        self.__colSep = "&##&\t"
        self.__rowSep = "$##$\n"
        #
        #
        self.__fTypeRow = "skip-max-width"
        self.__fTypeCol = "skip-max-width"
        #
        self.__warningAction = warnings
        dtf = DataTransformFactory(schemaDefAccessObj=self.__sD, filterType=self.__fTypeRow)
        self.__sdp = SchemaDefDataPrep(schemaDefAccessObj=self.__sD, dtObj=dtf, workPath=self.__cachePath, verbose=self.__verbose)
        self.__rpP = RepositoryProvider(cfgOb=self.__cfgOb, cachePath=self.__cachePath)
        #
        schemaName = self.__sD.getName()
        modulePathMap = self.__cfgOb.get("DICT_METHOD_HELPER_MODULE_PATH_MAP", sectionName=sectionName)
        dP = DictionaryApiProviderWrapper(self.__cachePath, cfgOb=self.__cfgOb, configName=sectionName, useCache=True)
        dictApi = dP.getApiByName(schemaName)
        rP = DictMethodResourceProvider(self.__cfgOb, cachePath=self.__cachePath, restoreUseStash=restoreUseStash, restoreUseGit=restoreUseGit, providerTypeExclude=providerTypeExclude)
        self.__dmh = DictMethodRunner(dictApi, modulePathMap=modulePathMap, resourceProvider=rP)

    def setWarning(self, action):
        if action in ["error", "ignore", "default"]:
            self.__warningAction = action
            return True
        else:
            self.__warningAction = "default"
            return False

    def setDelimiters(self, colSep=None, rowSep=None):
        """Set column and row delimiters for intermediate data files used for
        batch-file loading operations.
        """
        self.__colSep = colSep if colSep is not None else "&##&\t"
        self.__rowSep = rowSep if rowSep is not None else "$##$\n"
        return True

    def load(self, inputPathList=None, containerList=None, loadType="batch-file", deleteOpt=None, tableIdSkipD=None):
        """Load data for each table defined in the current schema definition object.
        Data are extracted from the input file list.

        Data source options:

          inputPathList = [<full path of target input file>, ....]

        or

          containerList = [ data container, ...]


        loadType  =  ['batch-file' | 'batch-insert']
        deleteOpt = 'selected' | 'all'

        tableIdSkipD - searchable container with tableIds to be skipped on loading -

        Loading is performed using the current database server connection.

        Intermediate data files for 'batch-file' loading are created in the current working path.

        Returns True for success or False otherwise.

        """
        tableIdSkipD = tableIdSkipD if tableIdSkipD is not None else {}
        if inputPathList is not None:
            cL = self.__rpP.getContainerList(inputPathList)
            #
            # Apply dynamic methods here -
            #
            for cA in cL:
                self.__dmh.apply(cA)
            tableDataDict, containerNameList = self.__sdp.process(cL)

        elif containerList is not None:
            tableDataDict, containerNameList = self.__sdp.process(containerList)
        #
        #
        if loadType in ["batch-file", "batch-file-append"]:
            append = True if loadType == "batch-file-append" else False
            exportList = self.__exportTdd(tableDataDict, colSep=self.__colSep, rowSep=self.__rowSep, append=append)
            for tableId, loadPath in exportList:
                if tableId in tableIdSkipD:
                    continue
                self.__batchFileImport(tableId, loadPath, sqlFilePath=None, containerNameList=containerNameList, deleteOpt=deleteOpt)
                if self.__cleanUp:
                    self.__cleanUpFile(loadPath)
            return True
        elif loadType == "batch-insert":
            for tableId, rowList in tableDataDict.items():
                if tableId in tableIdSkipD:
                    continue
                if deleteOpt in ["all", "selected"] or rowList:
                    self.__batchInsertImport(tableId, rowList=rowList, containerNameList=containerNameList, deleteOpt=deleteOpt)
            return True
        else:
            pass

        return False

    def __cleanUpFile(self, filePath):
        try:
            os.remove(filePath)
        except Exception:
            pass

    def makeLoadFilesMulti(self, dataList, procName, optionsD, workingDir):
        """Create a loadable data file for each table defined in the current schema
        definition object.   Data is extracted from the input file list.

        Load files are creating in the current working path.

        Return the containerNames for the input path list, and path list for load files that are created.

        """
        _ = workingDir
        try:
            pn = procName.split("-")[-1]
        except Exception:
            pn = procName

        exportFormat = optionsD["exportFormat"] if "exportFormat" in optionsD else "tdd"
        r1, r2 = self.makeLoadFiles(inputPathList=dataList, partName=pn, exportFormat=exportFormat)
        return dataList, r1, r2, []

    def makeLoadFiles(self, inputPathList, append=False, partName="1", exportFormat="tdd"):
        """Create a loadable data file for each table defined in the current schema
        definition object.   Data is extracted from the input file list.

        Load files are created in the current working path.

        Return the containerNames for the input path list, and path list for load files that are created.

        """
        cL = self.__rpP.getContainerList(inputPathList)
        for cA in cL:
            self.__dmh.apply(cA)
        tableDataDict, containerNameList = self.__sdp.process(cL)
        if exportFormat == "tdd":
            return containerNameList, self.__exportTdd(tableDataDict, colSep=self.__colSep, rowSep=self.__rowSep, append=append, partName=partName)
        elif exportFormat == "csv":
            return containerNameList, self.__exportCsv(tableDataDict, append=append, partName=partName)
        else:
            return [], []

    def __exportCsv(self, tableDict, append=False, partName="1"):
        """ """
        modeOpt = "a" if append else "w"

        exportList = []
        for tableId, rowList in tableDict.items():
            if not rowList:
                continue
            tObj = self.__sD.getSchemaObject(tableId)
            schemaAttributeIdList = tObj.getAttributeIdList()
            attributeNameList = tObj.getAttributeNameList()
            #
            fn = os.path.join(self.__workPath, tableId + "-" + partName + ".csv")
            with open(fn, modeOpt, newline="") as ofh:
                csvWriter = csv.writer(ofh)
                csvWriter.writerow(attributeNameList)
                for rD in rowList:
                    csvWriter.writerow([rD[aId] for aId in schemaAttributeIdList])

            exportList.append((tableId, fn))
        return exportList

    def __exportTdd(self, tableDict, colSep="&##&\t", rowSep="$##$\n", append=False, partName="1"):
        modeOpt = "a" if append else "w"

        exportList = []
        for tableId, rowList in tableDict.items():
            tObj = self.__sD.getSchemaObject(tableId)
            schemaAttributeIdList = tObj.getAttributeIdList()
            #
            if rowList:
                fn = os.path.join(self.__workPath, tableId + "-" + partName + ".tdd")
                ofh = open(fn, modeOpt)
                for rD in rowList:
                    # logger.info("%r" % colSep.join([str(rD[aId]) for aId in schemaAttributeIdList]))
                    ofh.write("%s%s" % (colSep.join([str(rD[aId]) for aId in schemaAttributeIdList]), rowSep))
                ofh.close()
                exportList.append((tableId, fn))
        return exportList

    def loadBatchFiles(self, loadList=None, containerNameList=None, deleteOpt=None):
        """Load data for each table defined in the current schema definition object using

        Data source options:

          loadList = [(tableId, <full path of load file), ....]
          containerNameList = [ data namecontainer, ...]

        deleteOpt = 'selected' | 'all','truncate'

        Loading is performed using the current database server connection.

        Returns True for success or False otherwise.

        """
        #
        startTime = time.time()
        for tableId, loadPath in loadList:
            ok = self.__batchFileImport(tableId, loadPath, sqlFilePath=None, containerNameList=containerNameList, deleteOpt=deleteOpt)
            if not ok:
                break
            if self.__cleanUp:
                self.__cleanUpFile(loadPath)
        #
        endTime = time.time()
        logger.debug("Completed with status %r at %s (%.3f seconds)\n", ok, time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - startTime)
        return ok

    def delete(self, tableId, containerNameList=None, deleteOpt="all"):
        #
        startTime = time.time()
        sqlCommandList = self.__getSqlDeleteList(tableId, containerNameList=containerNameList, deleteOpt=deleteOpt)

        myQ = MyDbQuery(dbcon=self.__dbCon, verbose=self.__verbose)
        myQ.setWarning(self.__warningAction)
        ret = myQ.sqlCommand(sqlCommandList=sqlCommandList)
        #
        #
        endTime = time.time()

        logger.debug("Delete table %s server returns %r\n", tableId, ret)
        logger.debug("Completed at %s (%.3f seconds)\n", time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - startTime)
        return ret

    def __getSqlDeleteList(self, tableId, containerNameList=None, deleteOpt="all"):
        """Return the SQL delete commands for the input table and container name list."""
        databaseName = self.__sD.getDatabaseName()
        sqlGen = SqlGenAdmin(self.__verbose)

        databaseName = self.__sD.getDatabaseName()
        tableDefObj = self.__sD.getSchemaObject(tableId)
        tableName = tableDefObj.getName()

        sqlDeleteList = []
        if deleteOpt in ["selected", "delete"] and containerNameList is not None:
            deleteAttributeName = tableDefObj.getDeleteAttributeName()
            sqlDeleteList = sqlGen.deleteFromListSQL(databaseName, tableName, deleteAttributeName, containerNameList, chunkSize=50)
        elif deleteOpt in ["all", "truncate"]:
            sqlDeleteList = [sqlGen.truncateTableSQL(databaseName, tableName)]

        logger.debug("Delete SQL for %s : %r\n", tableId, sqlDeleteList)
        return sqlDeleteList

    def __batchFileImport(self, tableId, tableLoadPath, sqlFilePath=None, containerNameList=None, deleteOpt="all"):
        """Batch load the input table using data in the input loadable data file.

        if sqlFilePath is provided then any generated SQL commands are preserved in this file.

        deleteOpt None|'selected'| 'all' or 'truncate'
        """
        startTime = time.time()
        databaseName = self.__sD.getDatabaseName()
        sqlGen = SqlGenAdmin(self.__verbose)

        databaseName = self.__sD.getDatabaseName()
        tableDefObj = self.__sD.getSchemaObject(tableId)
        # tableName = tableDefObj.getName()

        #
        if deleteOpt:
            sqlCommandList = self.__getSqlDeleteList(tableId, containerNameList=containerNameList, deleteOpt=deleteOpt)
        else:
            sqlCommandList = []

        if os.access(tableLoadPath, os.R_OK):
            tableDefObj = self.__sD.getSchemaObject(tableId)

            sqlCommandList.append(sqlGen.importTable(databaseName, tableDefObj, importPath=tableLoadPath))

            if self.__verbose:
                logger.debug("SQL import command\n%s\n", sqlCommandList)
            #

        if sqlFilePath is not None:
            try:
                with open(sqlFilePath, "w") as ofh:
                    ofh.write("%s" % "\n".join(sqlCommandList))
            except Exception:
                pass
        #
        myQ = MyDbQuery(dbcon=self.__dbCon, verbose=self.__verbose)
        myQ.setWarning(self.__warningAction)
        ret = myQ.sqlCommand(sqlCommandList=sqlCommandList)
        #
        #
        endTime = time.time()
        logger.debug("Table %s server returns %r\n", tableId, ret)
        logger.debug("Completed at %s (%.3f seconds)\n", time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - startTime)
        return ret

    def loadBatchData(self, tableId, rowList=None, containerNameList=None, deleteOpt="selected"):
        return self.__batchInsertImport(tableId, rowList=rowList, containerNameList=containerNameList, deleteOpt=deleteOpt)

    def __batchInsertImport(self, tableId, rowList=None, containerNameList=None, deleteOpt="selected"):
        """Load the input table using batch inserts of the input list of dictionaries (i.e. d[attributeId]=value).

        The containerNameList corresponding to the data within loadable data in rowList can be provided
        if 'selected' deletions are to performed prior to the the batch data inserts.

        deleteOpt = ['selected','all'] where 'selected' deletes rows corresponding to the input container
                    list before insert.   The 'all' options truncates the table prior to insert.

                    Deletions are performed in the absence of loadable data.

        """
        startTime = time.time()

        myQ = MyDbQuery(dbcon=self.__dbCon, verbose=self.__verbose)
        myQ.setWarning(self.__warningAction)
        sqlGen = SqlGenAdmin(self.__verbose)
        #
        databaseName = self.__sD.getDatabaseName()
        tableDefObj = self.__sD.getSchemaObject(tableId)
        tableName = tableDefObj.getName()
        tableAttributeIdList = tableDefObj.getAttributeIdList()
        tableAttributeNameList = tableDefObj.getAttributeNameList()
        #
        sqlDeleteList = None
        if deleteOpt in ["selected", "delete"] and containerNameList is not None:
            deleteAttributeName = tableDefObj.getDeleteAttributeName()
            sqlDeleteList = sqlGen.deleteFromListSQL(databaseName, tableName, deleteAttributeName, containerNameList, chunkSize=10)
            if self.__verbose:
                logger.debug("Delete SQL for %s : %r\n", tableId, sqlDeleteList)
        elif deleteOpt in ["all", "truncate"]:
            sqlDeleteList = [sqlGen.truncateTableSQL(databaseName, tableName)]

        sqlInsertList = []
        for row in rowList:
            vList = []
            aList = []
            for tid, nm in zip(tableAttributeIdList, tableAttributeNameList):
                # if len(row[id]) > 0 and row[id] != r'\N':
                if row[tid] is not None and row[tid] != r"\N":
                    vList.append(row[tid])
                    aList.append(nm)
            sqlInsertList.append((sqlGen.insertTemplateSQL(databaseName, tableName, aList), vList))

        ret = myQ.sqlBatchTemplateCommand(sqlInsertList, prependSqlList=sqlDeleteList)
        if ret:
            logger.debug("Batch insert completed for table %s rows %d\n", tableName, len(sqlInsertList))
        else:
            logger.error("Batch insert fails for table %s length %d\n", tableName, len(sqlInsertList))

        endTime = time.time()
        if self.__verbose:
            logger.debug("Completed at %s (%.3f seconds)\n", time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - startTime)

        return ret

    def __deleteFromTable(self, tableIdList, deleteValue):
        """Delete data from the input table list where the schema table delete attribute
        has the input value "deleteValue".

        """
        databaseName = self.__sD.getDatabaseName()
        sqlList = []
        sqlGen = SqlGenAdmin(self.__verbose)
        for tableId in tableIdList:
            tableName = self.__sD.getSchemaName(tableId)
            tableDefObj = self.__sD.getSchemaObject(tableId)
            atName = tableDefObj.getDeleteAttributeName()
            sqlTemp = sqlGen.deleteTemplateSQL(databaseName, tableName, [atName])
            sqlList.append(sqlTemp % deleteValue)
        #
        return sqlList



##
# File: RepoHoldingsDataPrep.py
# Author:  J. Westbrook
# Date:  12-Jul-2018
#
# Note:  This module uses well sandbox and repository defined file names
#        within the configuration defined RCSB exchange sandbox path.
#
# Update:
# 16-Jul-2018 jdw adjust naming to current sandbox conventions.
# 30-Oct-2018 jdw naming to conform to dictioanry and schema conventions.
#  8-Oct-2018 jdw adjustments for greater schema compliance and uniformity
#                 provide alternative date assignment for JSON and mongo validation
# 10-Oct-2018 jdw Add getHoldingsTransferred() for homology models transfered to Model Archive
# 28-Oct-2018 jdw update with semi-colon separated author lists in status and theoretical model files
# 25-Nov-2018 jdw add sequence/pdb_seq_prerelease.fasta and  _rcsb_repository_holdings_prerelease.seq_one_letter_code
# 27-Nov-2018 jdw update rcsb_repository_holdings_current for all entry content types
# 29-Nov-2018 jdw Add support for NMR restraint versions.
# 30-Nov-3018 jdw explicitly filter obsolete entries from current holdings
# 13-Dec-2018 jdw Adjust logic for reporting assembly format availibility
#  5-Feb-2020 jdw Drop superseded entries from the removed entry candidate list.
#                 Avoid overlap between current and removed/unreleased entries.
# 30-Apr-2020 jdw new NMR content types and support for config option RCSB_EDMAP_LIST_PATH
# 23-Oct-2020 jdw add getHoldingsCombined()
##

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"


import logging
import os

import dateutil.parser
from rcsb.utils.io.MarshalUtil import MarshalUtil

logger = logging.getLogger(__name__)


class RepoHoldingsDataPrep(object):
    """Consolidate legacy data describing repository content updates and repository entry status."""

    def __init__(self, **kwargs):
        self.__cfgOb = kwargs.get("cfgOb", None)
        self.__cachePath = kwargs.get("cachePath", None)
        self.__sandboxPath = kwargs.get("sandboxPath", None)
        self.__filterType = kwargs.get("filterType", "")
        self.__assignDates = "assign-dates" in self.__filterType
        #
        self.__mU = MarshalUtil(workPath=self.__cachePath)
        self.__currentCacheD = None
        #

    def getHoldingsCombinedEntry(self, updateId, dirPath=None):
        dList = []
        retD = self.__getHoldingsCombined(dirPath=dirPath)
        for entryId, qD in retD.items():
            tD = {"rcsb_id": entryId, "entry_id": entryId, "update_id": updateId}
            rD = {
                "rcsb_id": entryId,
                "rcsb_repository_holdings_combined_entry_container_identifiers": tD,
                "rcsb_repository_holdings_combined": qD,
            }
            dList.append(rD)
        return dList

    def __getHoldingsCombined(self, dirPath=None):
        retD = {}
        dirPath = dirPath if dirPath else self.__sandboxPath
        currentD = self.__currentCacheD if self.__currentCacheD else self.__getHoldingsCurrent(dirPath=dirPath)
        for entryId, tD in currentD.items():
            retD[entryId] = {"status": "CURRENT", "status_code": "REL"}
        logger.debug("Released entries %d", len(retD))
        #
        unRelD = self.__getHoldingsUnreleased(dirPath=dirPath)
        # logger.info("@@@ unRelD %r", unRelD)
        for entryId, tD in unRelD.items():
            if entryId not in retD and tD["status_code"] in ["AUCO", "AUTH", "HOLD", "HPUB", "POLC", "PROC", "REFI", "REPL", "WAIT", "WDRN"]:
                retD[entryId] = {"status": "UNRELEASED", "status_code": tD["status_code"]}
        logger.debug("Released & unreleased entries %d", len(retD))
        #
        trfD, _ = self.__getHoldingsTransferred(dirPath=dirPath)
        for entryId, tD in trfD.items():
            if entryId not in retD and tD["status_code"] in ["TRSF"]:
                retD[entryId] = {"status": "REMOVED", "status_code": tD["status_code"]}
        #
        logger.debug("Released & unreleased & transferred entries %d", len(retD))
        #
        rmvD, _, replacesD = self.__getHoldingsRemoved(dirPath=dirPath)
        #
        # for entryId in rmvD:
        #    if entryId not in retD:
        #        retD[entryId] = {"status": "REMOVED", "status_code": "OBS"}
        #
        replacedByD = {}
        for entryId, tD in replacesD.items():
            for sId in tD["id_codes_superseded"]:
                replacedByD[sId.strip().upper()] = entryId.strip().upper()
        #
        logger.info("replacedbyD (%d) rmvD (%d) currentD (%d) retD (%d)", len(replacedByD), len(rmvD), len(currentD), len(retD))
        for entryId in rmvD:
            if entryId in currentD:
                continue
            tId = entryId
            if tId in replacedByD:
                if tId == replacedByD[tId]:
                    logger.info("Inconsistent obsolete entry info for %r", tId)
                while tId in replacedByD and tId != replacedByD[tId]:
                    # logger.debug("tId %r replacedByD[tId] %r", tId, replacedByD[tId])
                    tId = replacedByD[tId]
                if tId in currentD:
                    retD[entryId] = {"status": "REMOVED", "status_code": "OBS", "id_code_replaced_by_latest": tId}
                else:
                    logger.debug("%r missing replacedby entry %r", entryId, tId)
            else:
                retD[entryId] = {"status": "REMOVED", "status_code": "OBS"}
        #
        logger.debug("Released & unreleased & transferred & removed entries %d", len(retD))
        return retD

    def getHoldingsCurrentEntry(self, updateId, dirPath=None):
        dList = []
        retD = self.__currentCacheD if self.__currentCacheD else self.__getHoldingsCurrent(dirPath=dirPath)
        self.__currentCacheD = retD
        for entryId, qD in retD.items():
            tD = (
                {"rcsb_id": entryId, "entry_id": entryId, "update_id": updateId, "assembly_ids": qD["assembly_ids"]}
                if "assembly_ids" in qD
                else {"rcsb_id": entryId, "entry_id": entryId, "update_id": updateId}
            )
            rD = {
                "rcsb_id": entryId,
                "rcsb_repository_holdings_current_entry_container_identifiers": tD,
                "rcsb_repository_holdings_current": {"repository_content_types": qD["repository_content_types"]},
            }
            dList.append(rD)
        return dList

    def getHoldingsUpdateEntry(self, updateId, dirPath=None):
        dList = []
        retD = self.__getHoldingsUpdate(dirPath=dirPath)
        for entryId, qD in retD.items():
            tD = {"rcsb_id": entryId, "entry_id": entryId, "update_id": updateId}
            rD = {
                "rcsb_id": entryId,
                "rcsb_repository_holdings_update_entry_container_identifiers": tD,
                "rcsb_repository_holdings_update": qD,
            }
            dList.append(rD)
        return dList

    def getHoldingsUnreleasedEntry(self, updateId, dirPath=None):
        dList = []
        retD = self.__getHoldingsUnreleased(dirPath=dirPath)
        prD = self.__getHoldingsPrerelease(dirPath=dirPath)
        currentD = self.__currentCacheD if self.__currentCacheD else self.__getHoldingsCurrent(dirPath=dirPath)
        self.__currentCacheD = currentD
        for entryId, qD in retD.items():
            if entryId in currentD:
                continue
            rD = {"rcsb_id": entryId}
            rD["rcsb_repository_holdings_unreleased_entry_container_identifiers"] = {"rcsb_id": entryId, "entry_id": entryId, "update_id": updateId}
            if entryId in prD:
                rD["rcsb_repository_holdings_prerelease"] = prD[entryId]
                qD["prerelease_sequence_available_flag"] = "Y"
            else:
                qD["prerelease_sequence_available_flag"] = "N"
            rD["rcsb_repository_holdings_unreleased"] = qD
            #
            dList.append(rD)
        return dList

    def getHoldingsRemovedEntry(self, updateId, dirPath=None):
        dList = []
        rmvD, aaD, spsD = self.__getHoldingsRemoved(dirPath=dirPath)
        trfD, insD = self.__getHoldingsTransferred(dirPath=dirPath)
        currentD = self.__currentCacheD if self.__currentCacheD else self.__getHoldingsCurrent(dirPath=dirPath)
        self.__currentCacheD = currentD
        #
        # Get the list of candidate keys for removed entries -
        #
        entryIdL = sorted(set(list(insD.keys()) + list(rmvD.keys())))
        for entryId in entryIdL:
            if entryId in currentD:
                continue
            rD = {"rcsb_id": entryId}
            rD["rcsb_repository_holdings_removed_entry_container_identifiers"] = {"rcsb_id": entryId, "entry_id": entryId, "update_id": updateId}
            #
            if entryId in rmvD:
                rD["rcsb_repository_holdings_removed"] = rmvD[entryId]
            if entryId in aaD:
                rD["rcsb_repository_holdings_removed_audit_author"] = aaD[entryId]
            if entryId in spsD:
                rD["rcsb_repository_holdings_superseded"] = spsD[entryId]
            if entryId in trfD:
                rD["rcsb_repository_holdings_transferred"] = trfD[entryId]
            if entryId in insD:
                rD["rcsb_repository_holdings_insilico_models"] = insD[entryId]
            dList.append(rD)
        return dList

    def __getHoldingsTransferred(self, dirPath=None):
        """Parse legacy lists defining the repository contents transferred to alternative repositories

        Args:
            updateId (str): update identifier (e.g. 2018_32)
            dirPath (str): directory path containing update list files
            **kwargs: unused

        Returns:
            (dict): dictionaries containing data for rcsb_repository_holdings_transferred
            (dict): dictionaries containing data for rcsb_repository_holdings_insilico_models

        Example input data:

        ma-czyyf : 262D - TITLE A THREE-DIMENSIONAL MODEL OF THE REV BINDING ELEMENT OF HIV- TITLE 2 1 DERIVED FROM ANALYSES OF IN VITRO SELECTED VARIANTS
        ma-cfqla : 163D - TITLE A THREE-DIMENSIONAL MODEL OF THE REV BINDING ELEMENT OF HIV- TITLE 2 1 DERIVED FROM ANALYSES OF IN VITRO SELECTED VARIANTS

        and -

        1DX2    REL 1999-12-16  2000-12-15  Tumour Targetting Human ...  Beiboer, S.H.W., Reurs, A., Roovers, R.C., Arends, J., Whitelegg, N.R.J., Rees, A.R., Hoogenboom, H.R.

        and -

        1APD    OBSLTE  1992-10-15      2APD
        1BU0    OBSLTE  1998-10-07      2BU0
        1CLJ    OBSLTE  1998-03-04      2CLJ
        1DU8    OBSLTE  2001-01-31      1GIE
        1I2J    OBSLTE  2001-01-06      1JA5

        """
        trsfD = {}
        insD = {}
        dirPath = dirPath if dirPath else self.__sandboxPath

        try:
            fp = os.path.join(dirPath, "status", "theoretical_model_obsolete.tsv")
            lineL = self.__mU.doImport(fp, "list")  # pylint: disable=no-member
            #
            obsDateD = {}
            obsIdD = {}
            for line in lineL:
                fields = line.split("\t")
                if len(fields) < 3:
                    continue
                entryId = str(fields[0]).strip().upper()
                obsDateD[entryId] = dateutil.parser.parse(fields[2]) if self.__assignDates else fields[2]
                if len(fields) > 3 and len(fields[3]) > 3:
                    obsIdD[entryId] = str(fields[3]).strip().upper()
            logger.debug("Read %d obsolete insilico id codes", len(obsDateD))
            # ---------  ---------  ---------  ---------  ---------  ---------  ---------
            fp = os.path.join(dirPath, "status", "model-archive-PDB-insilico-mapping.list")
            lineL = self.__mU.doImport(fp, "list")
            #
            trD = {}
            for line in lineL:
                fields = line.split(":")
                if len(fields) < 2:
                    continue
                entryId = str(fields[1]).strip().upper()[:4]
                maId = str(fields[0]).strip()
                trD[entryId] = maId
            logger.debug("Read %d model archive id codes", len(trD))
            #
            # ---------  ---------  ---------  ---------  ---------  ---------  ---------
            fp = os.path.join(dirPath, "status", "theoretical_model_v2.tsv")
            lineL = self.__mU.doImport(fp, "list")
            #
            logger.debug("Read %d insilico id codes", len(lineL))
            for line in lineL:
                fields = str(line).split("\t")
                if len(fields) < 6:
                    continue
                depDate = dateutil.parser.parse(fields[2]) if self.__assignDates else fields[2]
                relDate = None
                if len(fields[3]) >= 10 and not fields[3].startswith("0000"):
                    relDate = dateutil.parser.parse(fields[3]) if self.__assignDates else fields[3]

                statusCode = "TRSF" if fields[1] == "REL" else fields[1]

                entryId = str(fields[0]).upper()
                title = fields[4]
                #
                auditAuthors = [t.strip() for t in fields[5].split(";")]
                repId = None
                if entryId in trD:
                    repName = "Model Archive"
                    repId = trD[entryId]

                #
                dD = {
                    "status_code": statusCode,
                    "deposit_date": depDate,
                    "repository_content_types": ["coordinates"],
                    "title": title,
                    "audit_authors": auditAuthors,
                }
                #
                if relDate:
                    dD["release_date"] = relDate
                #
                if repId:
                    dD["remote_accession_code"] = repId
                    dD["remote_repository_name"] = repName
                if statusCode == "TRSF":
                    trsfD[entryId] = dD
                #
                #
                dD = {"status_code": statusCode, "deposit_date": depDate, "title": title, "audit_authors": auditAuthors}
                #
                if relDate:
                    dD["release_date"] = relDate
                #
                if entryId in obsDateD:
                    dD["remove_date"] = relDate
                #
                if entryId in obsIdD:
                    dD["id_codes_replaced_by"] = [obsIdD[entryId]]
                #
                insD[entryId] = dD
            #
            logger.info("Transferred entries %d - insilico models %d", len(trsfD), len(insD))
            #
        except Exception as e:
            logger.exception("Failing with %s", str(e))

        return trsfD, insD

    def __getHoldingsUpdate(self, dirPath=None):
        """Parse legacy lists defining the contents of the repository update

        Args:
            updateId (str): update identifier (e.g. 2018_32)
            dirPath (str): directory path containing update list files
            **kwargs: unused

        Returns:
            list: List of dictionaries containing rcsb_repository_holdings_update
        """
        retD = {}
        dirPath = dirPath if dirPath else self.__sandboxPath
        try:
            updateTypeList = ["added", "modified", "obsolete"]
            contentTypeList = ["entries", "mr", "cs", "sf", "nef", "nmr-str"]
            contentNameD = {
                "entries": "coordinates",
                "mr": "NMR restraints",
                "cs": "NMR chemical shifts",
                "sf": "structure factors",
                "nef": "Combined NMR data (NEF)",
                "nmr-str": "Combined NMR data (NMR-STAR)",
            }
            #
            uD = {}
            for updateType in updateTypeList:
                for contentType in contentTypeList:
                    fp = os.path.join(dirPath, "update-lists", updateType + "-" + contentType)
                    if not self.__mU.exists(fp):
                        continue
                    entryIdL = self.__mU.doImport(fp, "list")
                    #
                    for entryId in entryIdL:
                        entryId = entryId.strip().upper()
                        if entryId not in uD:
                            uD[entryId] = []
                        uD[entryId].append(contentNameD[contentType])
                for entryId in uD:
                    uType = "removed" if updateType == "obsolete" else updateType
                    # retD[entryId] = {"update_id": updateId, "entry_id": entryId, "update_type": uType, "repository_content_types": uD[entryId]}
                    retD[entryId] = {"update_type": uType, "repository_content_types": uD[entryId]}
            return retD
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return retD

    def __getHoldingsCurrent(self, dirPath=None):
        """Parse legacy lists defining the current contents of the repository update

        Args:
            updateId (str): update identifier (e.g. 2018_32)
            dirPath (str): directory path containing update list files
            **kwargs: unused

        Returns:
            list: List of dictionaries containing data for rcsb_repository_holdings_current
        """
        rD = {}
        retD = {}
        dirPath = dirPath if dirPath else self.__sandboxPath
        try:
            updateTypeList = ["all"]
            contentTypeList = ["pdb", "pdb-format", "mr", "cs", "sf", "nef", "nmr-str"]
            contentNameD = {
                "pdb": "coordinates",
                "pdb-format": "PDB format coordinates",
                "mr": "NMR restraints",
                "cs": "NMR chemical shifts",
                "sf": "structure factors",
                "nef": "Combined NMR data (NEF)",
                "nmr-str": "Combined NMR data (NMR-STAR)",
            }
            #
            tD = {}
            for updateType in updateTypeList:
                for contentType in contentTypeList:
                    fp = os.path.join(dirPath, "update-lists", updateType + "-" + contentType + "-list")
                    if not self.__mU.exists(fp):
                        continue
                    entryIdL = self.__mU.doImport(fp, "list")
                    #
                    for entryId in entryIdL:
                        entryId = entryId.strip().upper()
                        if entryId not in tD:
                            tD[entryId.upper()] = {}
                        tD[entryId.upper()][contentNameD[contentType]] = True
            #
            fp = os.path.join(dirPath, "status", "biounit_file_list.tsv")
            lines = self.__mU.doImport(fp, "list")
            assemD = {}
            for line in lines:
                fields = line.split("\t")
                entryId = fields[0].strip().upper()
                assemId = fields[1].strip()
                if entryId not in assemD:
                    assemD[entryId.upper()] = []
                assemD[entryId.upper()].append(assemId)
            #
            #
            fp = os.path.join(dirPath, "status", "pdb_bundle_index_list.tsv")
            bundleIdList = self.__mU.doImport(fp, "list")
            bundleD = {}
            for entryId in bundleIdList:
                bundleD[entryId.strip().upper()] = True
            #
            fp = os.path.join(dirPath, "status", "validation_report_list_new.tsv")
            vList = self.__mU.doImport(fp, "list")
            valD = {}
            valImageD = {}
            valCifD = {}
            for line in vList:
                fields = line.split("\t")
                entryId = fields[0].strip().upper()
                imageFlag = fields[1].strip().upper()
                valD[entryId] = True
                valImageD[entryId] = imageFlag == "Y"
                if len(fields) > 2:
                    valCifD[entryId] = fields[2].strip().upper() == "Y"
            #
            #
            fp = os.path.join(dirPath, "status", "entries_without_polymers.tsv")
            pList = self.__mU.doImport(fp, "list")
            pD = {}
            for entryId in pList:
                pD[entryId.strip().upper()] = False
            #
            #
            fp = os.path.join(dirPath, "status", "nmr_restraints_v2_list.tsv")
            nmrV2List = self.__mU.doImport(fp, "list")
            nmrV2D = {}
            for entryId in nmrV2List:
                nmrV2D[entryId.strip().upper()] = False
            #
            if self.__cfgOb:
                configName = self.__cfgOb.getDefaultSectionName()
                fp = self.__cfgOb.getPath("RCSB_EDMAP_LIST_PATH", sectionName=configName)
            else:
                fp = os.path.join(dirPath, "status", "edmaps.json")
            qD = self.__mU.doImport(fp, "json")
            edD = {}
            for entryId in qD:
                edD[entryId.upper()] = qD[entryId]
            #
            fp = os.path.join(dirPath, "status", "obsolete_entry.json_2")
            oL = self.__mU.doImport(fp, "json")
            obsD = {}
            for dD in oL:
                obsD[dD["entryId"].upper()] = True
            logger.info("Removed entry length %d", len(obsD))
            #
            #
            # Revise content types bundles and assemblies
            #
            for qId, dD in tD.items():
                entryId = qId.strip().upper()
                if entryId in obsD:
                    continue
                rD[entryId] = []
                if entryId in bundleD:
                    rD[entryId].append("entry PDB bundle")
                if "coordinates" in dD:
                    rD[entryId].append("entry mmCIF")
                    rD[entryId].append("entry PDBML")
                if "PDB format coordinates" in dD:
                    rD[entryId].append("entry PDB")
                if entryId in assemD:
                    if entryId in bundleD:
                        rD[entryId].append("assembly mmCIF")
                    else:
                        rD[entryId].append("assembly PDB")
                #
                for cType in dD:
                    if cType not in ["coordinates", "PDB format coordinates", "NMR restraints"]:
                        rD[entryId].append(cType)
                    if cType == "NMR restraints":
                        rD[entryId].append("NMR restraints V1")

                if entryId in nmrV2D:
                    rD[entryId].append("NMR restraints V2")
                #
                if entryId in valD:
                    rD[entryId].append("validation report")
                if entryId in valImageD and valImageD[entryId]:
                    rD[entryId].append("validation slider image")
                if entryId in valCifD and valCifD[entryId]:
                    rD[entryId].append("validation data mmCIF")
                if entryId in edD:
                    rD[entryId].append("2fo-fc Map")
                    rD[entryId].append("fo-fc Map")
                    rD[entryId].append("Map Coefficients")
                if entryId not in pD:
                    rD[entryId].append("FASTA sequence")
            #
            for entryId in rD:
                if entryId in assemD:
                    retD[entryId] = {"assembly_ids": assemD[entryId], "repository_content_types": rD[entryId]}
                else:
                    retD[entryId] = {"repository_content_types": rD[entryId]}
            return retD
        except Exception as e:
            logger.exception("Failing with %s", str(e))

        return retD

    def __getHoldingsUnreleased(self, dirPath=None):
        """Parse the legacy exchange status file containing details for unreleased entries:

        Args:
            updateId (str): update identifier (e.g. 2018_32)
            dirPath (str): directory path containing update list files
            **kwargs: unused

        Returns:
            list: List of dictionaries containing data for rcsb_repository_holdings_unreleased

        """
        retD = {}
        fields = []
        dirPath = dirPath if dirPath else self.__sandboxPath
        try:
            #
            fp = os.path.join(dirPath, "status", "status_v2.txt")
            lines = self.__mU.doImport(fp, "list")
            for line in lines:
                fields = line.split("\t")
                if len(fields) < 15:
                    continue
                entryId = fields[1]
                dD = {
                    "status_code": fields[2]
                    # 'sg_project_name': fields[14],
                    # 'sg_project_abbreviation_': fields[15]}
                }
                if fields[11] and fields[11].strip():
                    dD["title"] = fields[11]
                if fields[10] and fields[10].strip():
                    dD["audit_authors"] = [t.strip() for t in fields[10].split(";")]
                    # d['audit_authors'] = fields[10]
                if fields[12] and fields[12].strip():
                    dD["author_prerelease_sequence_status"] = str(fields[12]).strip().replace("REALEASE", "RELEASE")
                dTupL = [
                    ("deposit_date", 3),
                    ("deposit_date_coordinates", 4),
                    ("deposit_date_structure_factors", 5),
                    ("hold_date_structure_factors", 6),
                    ("deposit_date_nmr_restraints", 7),
                    ("hold_date_nmr_restraints", 8),
                    ("release_date", 9),
                    ("hold_date_coordinates", 13),
                ]
                for dTup in dTupL:
                    fN = dTup[1]
                    if fields[fN] and len(fields[fN]) >= 4:
                        dD[dTup[0]] = dateutil.parser.parse(fields[fN]) if self.__assignDates else fields[fN]
                #
                retD[entryId] = {k: v for k, v in dD.items() if v}
        except Exception as e:
            logger.error("Fields: %r", fields)
            logger.exception("Failing with %s", str(e))

        return retD

    def __getHoldingsRemoved(self, dirPath=None):
        """Parse the legacy exchange file containing details of removed entries:

            {
                "entryId": "125D",
                "obsoletedDate": "1998-04-15",
                "title": "SOLUTION STRUCTURE OF THE DNA-BINDING DOMAIN OF CD=2=-GAL4 FROM S. CEREVISIAE",
                "details": "",
                "depositionAuthors": [
                    "Baleja, J.D.",
                    "Wagner, G."
                ],
                "depositionDate": "1993-05-05",
                "releaseDate": "1994-01-31",
                "obsoletedBy": [
                    "1AW6"
                ],
                "content_type": [
                    "entry mmCIF",
                    "entry PDB",
                    "entry PDBML",
                    "structure factors"
                ]},

        Returns;
            (dict) : dictionaries for rcsb_repository_holdings_removed
            (dict) : dictionaries for rcsb_repository_holdings_removed_audit_authors
            (dict) : dictionaries for rcsb_repository_holdings_superseded

        """
        # rcsb_repository_holdings_removed
        rL1D = {}
        # rcsb_repository_holdings_removed_audit_authors
        rL2D = {}
        # rcsb_repository_holdings_superseded
        rL3D = {}
        #
        sD = {}
        dirPath = dirPath if dirPath else self.__sandboxPath
        try:
            fp = os.path.join(dirPath, "status", "obsolete_entry.json_2")
            dD = self.__mU.doImport(fp, "json")
            for dT in dD:
                # ---
                ctL = dT["content_type"] if "content_type" in dT else []
                # ---
                rbL = dT["obsoletedBy"] if "obsoletedBy" in dT else []
                d1 = {"title": dT["title"], "details": dT["details"], "audit_authors": dT["depositionAuthors"]}
                if rbL:
                    d1["id_codes_replaced_by"] = [t.upper() for t in rbL]
                if ctL:
                    d1["repository_content_types"] = ctL

                dTupL = [("deposit_date", "depositionDate"), ("remove_date", "obsoletedDate"), ("release_date", "releaseDate")]
                for dTup in dTupL:
                    fN = dTup[1]
                    if dT[fN] and len(dT[fN]) > 4:
                        d1[dTup[0]] = dateutil.parser.parse(dT[fN]) if self.__assignDates else dT[fN]

                rL1D[dT["entryId"]] = {k: v for k, v in d1.items() if v}
                #
                for ii, author in enumerate(dT["depositionAuthors"]):
                    d2 = {"ordinal_id": ii + 1, "audit_author": author}
                    rL2D.setdefault(dT["entryId"], []).append(d2)
                if "obsoletedBy" in dT:
                    for pdbId in dT["obsoletedBy"]:
                        if pdbId not in sD:
                            sD[pdbId] = []
                        sD[pdbId].append(dT["entryId"])
            #
            for pdbId in sD:
                if sD[pdbId]:
                    rL3D[pdbId] = {"id_codes_superseded": sD[pdbId]}

            logger.debug("Computed data lengths  %d %d %d", len(rL1D), len(rL2D), len(rL3D))
        except Exception as e:
            logger.exception("Failing with %s", str(e))

        return rL1D, rL2D, rL3D

    def __getHoldingsPrerelease(self, dirPath=None):
        """Parse the legacy exchange status file containing prerelease sequence data.

        Args:
            updateId (str): update identifier (e.g. 2018_32)
            dirPath (str): directory path containing update list files
            **kwargs: unused

        Returns:
            list: List of dictionaries containing data for rcsb_repository_holdings_prerelease

        >6I99 Entity 1
        HHHHHHENLYFQGELKREEITLLKELGSGQFGVVKLGKWKGQYDVAVKMIKEG....
        >6JKE Entity 1
        GRVTNQLQYLHKVVMKALWKHQFAWPFRQPVDAVKLGLPDYHKIIKQPMDMGTI....

        """
        retD = {}
        fields = []
        dirPath = dirPath if dirPath else self.__sandboxPath
        try:
            # Get prerelease sequence data
            fp = os.path.join(dirPath, "sequence", "pdb_seq_prerelease.fasta")
            sD = self.__mU.doImport(fp, "fasta", commentStyle="prerelease")
            seqD = {}
            for sid in sD:
                fields = sid.split("_")
                entryId = str(fields[0]).upper()
                entityId = str(fields[1])
                if entryId not in seqD:
                    seqD[entryId] = []
                seqD[entryId].append((entityId, sD[sid]["sequence"]))
            logger.debug("Loaded prerelease sequences for %d entries", len(seqD))
            #
            for entryId, seqTupL in seqD.items():
                # dD = {"seq_one_letter_code": seqL}
                logger.debug("Adding prerelease sequences for %s", entryId)
                for entityId, seqS in seqTupL:
                    if not seqS:
                        continue
                    retD.setdefault(entryId, []).append({"entity_id": entityId, "seq_one_letter_code": seqS})
                #
                # retD[entryId] = {k: v for k, vTup in dD.items() if vTup[1]}
        except Exception as e:
            logger.error("Fields: %r", fields)
            logger.exception("Failing with %s", str(e))

        return retD



##
# File: RepoLoadExec.py
# Date: 15-Mar-2018  jdw
#
#  Execution wrapper  --  repository database loading utilities --
#  Updates:
#
#    21-Mar-2018 - jdw added content filters and separate collection for Bird chemical components
#    22-May-2018 - jdw add replacment load type, add options for input file paths
#    24-Mar-2018 - jdw add split collections for entries (preiliminary)
#    24-Mar-2018 - jdw add option to preserve paths from automatic repo scan
#    27-Mar-2018 - jdw update configuration handling and add support for mocking repository paths
#     4-Apr-2018 - jdw added option to prune documents to a size limit
#     3-Jul-2018 - jdw specialize for repository loading.
#    14-Jul-2018 - jdw add loading of separate
#    25-Jul-2018 - jdw die on input file path list processing error
#    20-Aug-2018 - jdw add load_pdbx_core load option
#     9-Sep-2018 - jdw expose --schema_level option
#     3-Dec-2018 - jdw add options to load specific core collections.
#    12-Dec-2018 - jdw add core_entity_monomer collection support
#    13-Dec-2018 - jdw add I/HM schema support
##
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import argparse
import logging
import os
import sys

from rcsb.utils.dictionary.DictMethodResourceProvider import DictMethodResourceProvider
from rcsb.db.mongo.DocumentLoader import DocumentLoader
from rcsb.db.mongo.PdbxLoader import PdbxLoader
from rcsb.utils.config.ConfigUtil import ConfigUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil

HERE = os.path.abspath(os.path.dirname(__file__))
TOPDIR = os.path.dirname(os.path.dirname(os.path.dirname(HERE)))

logging.basicConfig(level=logging.INFO, format="%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s")
logger = logging.getLogger()


def loadStatus(statusList, cfgOb, cachePath, readBackCheck=True):
    sectionName = "data_exchange_configuration"
    dl = DocumentLoader(cfgOb, cachePath, "MONGO_DB", numProc=2, chunkSize=2, documentLimit=None, verbose=False, readBackCheck=readBackCheck)
    #
    databaseName = cfgOb.get("DATABASE_NAME", sectionName=sectionName)
    # collectionVersion = cfgOb.get('COLLECTION_VERSION_STRING', sectionName=sectionName)
    collectionName = cfgOb.get("COLLECTION_UPDATE_STATUS", sectionName=sectionName)
    ok = dl.load(databaseName, collectionName, loadType="append", documentList=statusList, indexAttributeList=["update_id", "database_name", "object_name"], keyNames=None)
    return ok


def buildResourceCache(cfgOb, configName, cachePath, rebuildCache=False):
    """Generate and cache resource dependencies."""
    ret = False
    try:
        rp = DictMethodResourceProvider(cfgOb, configName=configName, cachePath=cachePath, restoreUseStash=True, restoreUseGit=True, providerTypeExclude=None)
        ret = rp.cacheResources(useCache=not rebuildCache, doBackup=False)
    except Exception as e:
        logger.exception("Failing with %s", str(e))
    return ret


def main():
    parser = argparse.ArgumentParser()
    #
    defaultConfigName = "site_info_configuration"
    #
    parser.add_argument("--full", default=False, action="store_true", help="Fresh full load in a new tables/collections")
    parser.add_argument("--replace", default=False, action="store_true", help="Load with replacement in an existing table/collection (default)")
    #
    parser.add_argument("--load_chem_comp_ref", default=False, action="store_true", help="Load Chemical Component reference definitions (public subset)")
    parser.add_argument("--load_chem_comp_core_ref", default=False, action="store_true", help="Load Chemical Component Core reference definitions (public subset)")
    parser.add_argument("--load_bird_chem_comp_ref", default=False, action="store_true", help="Load Bird Chemical Component reference definitions (public subset)")
    parser.add_argument("--load_bird_chem_comp_core_ref", default=False, action="store_true", help="Load Bird Chemical Component Core reference definitions (public subset)")
    parser.add_argument("--load_bird_ref", default=False, action="store_true", help="Load Bird reference definitions (public subset)")
    parser.add_argument("--load_bird_family_ref", default=False, action="store_true", help="Load Bird Family reference definitions (public subset)")
    parser.add_argument("--load_entry_data", default=False, action="store_true", help="Load PDBx entry data (current released subset)")
    parser.add_argument("--load_pdbx_core", default=False, action="store_true", help="Load all PDBx core collections (current released subset)")
    parser.add_argument("--load_pdbx_core_merge", default=False, action="store_true", help="Load all PDBx core collections with merged content (current released subset)")
    #
    parser.add_argument("--load_pdbx_core_entry", default=False, action="store_true", help="Load PDBx core entry (current released subset)")
    parser.add_argument("--load_pdbx_core_entity", default=False, action="store_true", help="Load PDBx core entity (current released subset)")
    parser.add_argument("--load_pdbx_core_entity_monomer", default=False, action="store_true", help="Load PDBx core entity monomer (current released subset)")
    parser.add_argument("--load_pdbx_core_assembly", default=False, action="store_true", help="Load PDBx core assembly (current released subset)")
    parser.add_argument("--load_ihm_dev", default=False, action="store_true", help="Load I/HM DEV model data (current released subset)")
    #
    parser.add_argument("--config_path", default=None, help="Path to configuration options file")
    parser.add_argument("--config_name", default=defaultConfigName, help="Configuration section name")

    parser.add_argument("--db_type", default="mongo", help="Database server type (default=mongo)")

    parser.add_argument(
        "--document_style",
        default="rowwise_by_name_with_cardinality",
        help="Document organization (rowwise_by_name_with_cardinality|rowwise_by_name|columnwise_by_name|rowwise_by_id|rowwise_no_name",
    )
    parser.add_argument("--read_back_check", default=False, action="store_true", help="Perform read back check on all documents")
    parser.add_argument("--schema_level", default=None, help="Schema validation level (full|min default=None)")
    #
    parser.add_argument("--load_file_list_path", default=None, help="Input file containing load file path list (override automatic repository scan)")
    parser.add_argument("--fail_file_list_path", default=None, help="Output file containing file paths that fail to load")
    parser.add_argument("--save_file_list_path", default=None, help="Save repo file paths from automatic file system scan in this path")

    parser.add_argument("--num_proc", default=2, help="Number of processes to execute (default=2)")
    parser.add_argument("--chunk_size", default=10, help="Number of files loaded per process")
    parser.add_argument("--file_limit", default=None, help="Load file limit for testing")
    parser.add_argument("--prune_document_size", default=None, help="Prune large documents to this size limit (MB)")
    parser.add_argument("--debug", default=False, action="store_true", help="Turn on verbose logging")
    parser.add_argument("--mock", default=False, action="store_true", help="Use MOCK repository configuration for testing")
    parser.add_argument("--cache_path", default=None, help="Cache path for resource files")
    parser.add_argument("--rebuild_cache", default=False, action="store_true", help="Rebuild cached resource files")
    parser.add_argument("--rebuild_schema", default=False, action="store_true", help="Rebuild schema on-the-fly if not cached")
    parser.add_argument("--vrpt_repo_path", default=None, help="Path to validation report repository")
    args = parser.parse_args()
    #
    debugFlag = args.debug
    if debugFlag:
        logger.setLevel(logging.DEBUG)
    # ----------------------- - ----------------------- - ----------------------- - ----------------------- - ----------------------- -
    #                                       Configuration Details
    configPath = args.config_path
    configName = args.config_name
    if not configPath:
        configPath = os.getenv("DBLOAD_CONFIG_PATH", None)
    try:
        if os.access(configPath, os.R_OK):
            os.environ["DBLOAD_CONFIG_PATH"] = configPath
            logger.info("Using configuation path %s (%s)", configPath, configName)
        else:
            logger.error("Missing or access issue with config file %r", configPath)
            exit(1)
        mockTopPath = os.path.join(TOPDIR, "rcsb", "mock-data") if args.mock else None
        cfgOb = ConfigUtil(configPath=configPath, defaultSectionName=defaultConfigName, mockTopPath=mockTopPath)
        if configName != defaultConfigName:
            cfgOb.replaceSectionName(defaultConfigName, configName)

        #
        if args.vrpt_repo_path:
            vrptPath = args.vrpt_repo_path
            if not os.access(vrptPath, os.R_OK):
                logger.error("Unreadable validation report repository path %r", vrptPath)
            envName = cfgOb.get("VRPT_REPO_PATH_ENV", sectionName=configName)
            os.environ[envName] = vrptPath
            logger.info("Using alternate validation report path %s", os.getenv(envName))

    except Exception as e:
        logger.error("Missing or access issue with config file %r with %s", configPath, str(e))
        exit(1)

    #
    try:
        readBackCheck = args.read_back_check
        numProc = int(args.num_proc)
        chunkSize = int(args.chunk_size)
        fileLimit = int(args.file_limit) if args.file_limit else None
        failedFilePath = args.fail_file_list_path
        fPath = args.load_file_list_path
        schemaLevel = args.schema_level if args.schema_level in ["min", "full", "minimum"] else None
        loadType = "full" if args.full else "replace"
        loadType = "replace" if args.replace else "full"
        saveInputFileListPath = args.save_file_list_path
        pruneDocumentSize = float(args.prune_document_size) if args.prune_document_size else None
        cachePath = args.cache_path if args.cache_path else "."
        cachePath = os.path.abspath(cachePath)
        rebuildCache = args.rebuild_cache if args.rebuild_cache else False
        rebuildSchemaFlag = args.rebuild_schema if args.rebuild_schema else False
        if args.document_style not in ["rowwise_by_name", "rowwise_by_name_with_cardinality", "columnwise_by_name", "rowwise_by_id", "rowwise_no_name"]:
            logger.error("Unsupported document style %s", args.document_style)
        if args.db_type != "mongo":
            logger.error("Unsupported database server type %s", args.db_type)
    except Exception as e:
        logger.exception("Argument processing problem %s", str(e))
        parser.print_help(sys.stderr)
        exit(1)

    # ----------------------- - ----------------------- - ----------------------- - ----------------------- - ----------------------- -
    #  Rebuild or check resource cache
    okS = True
    ok = buildResourceCache(cfgOb, configName, cachePath, rebuildCache=rebuildCache)
    if not ok:
        logger.error("Cache rebuild or check failure (rebuild %r) %r", rebuildCache, cachePath)
        exit(1)

    # ----------------------- - ----------------------- - ----------------------- - ----------------------- - ----------------------- -
    # Read any input path lists -
    #
    inputPathList = None
    if fPath:
        mu = MarshalUtil(workPath=cachePath)
        inputPathList = mu.doImport(fPath, fmt="list")
        if not inputPathList:
            logger.error("Missing or empty input file path list %s", fPath)
            exit(1)
    #
    ##
    if args.db_type == "mongo":
        mw = PdbxLoader(
            cfgOb,
            cachePath,
            resourceName="MONGO_DB",
            numProc=numProc,
            chunkSize=chunkSize,
            fileLimit=fileLimit,
            verbose=debugFlag,
            readBackCheck=readBackCheck,
            rebuildSchemaFlag=rebuildSchemaFlag,
        )

        if args.load_chem_comp_ref:
            ok = mw.load(
                "chem_comp",
                loadType=loadType,
                inputPathList=inputPathList,
                styleType=args.document_style,
                dataSelectors=["PUBLIC_RELEASE"],
                failedFilePath=failedFilePath,
                saveInputFileListPath=saveInputFileListPath,
                pruneDocumentSize=pruneDocumentSize,
                validationLevel=schemaLevel,
            )
            okS = loadStatus(mw.getLoadStatus(), cfgOb, cachePath, readBackCheck=readBackCheck)

        if args.load_chem_comp_core_ref:
            ok = mw.load(
                "chem_comp_core",
                loadType=loadType,
                inputPathList=inputPathList,
                styleType=args.document_style,
                dataSelectors=["PUBLIC_RELEASE"],
                failedFilePath=failedFilePath,
                saveInputFileListPath=saveInputFileListPath,
                pruneDocumentSize=pruneDocumentSize,
                validationLevel=schemaLevel,
            )
            okS = loadStatus(mw.getLoadStatus(), cfgOb, cachePath, readBackCheck=readBackCheck)

        if args.load_bird_chem_comp_ref:
            ok = mw.load(
                "bird_chem_comp",
                loadType=loadType,
                inputPathList=inputPathList,
                styleType=args.document_style,
                dataSelectors=["PUBLIC_RELEASE"],
                failedFilePath=failedFilePath,
                saveInputFileListPath=saveInputFileListPath,
                pruneDocumentSize=pruneDocumentSize,
                validationLevel=schemaLevel,
            )
            okS = loadStatus(mw.getLoadStatus(), cfgOb, cachePath, readBackCheck=readBackCheck)

        if args.load_bird_chem_comp_core_ref:
            ok = mw.load(
                "bird_chem_comp_core",
                loadType=loadType,
                inputPathList=inputPathList,
                styleType=args.document_style,
                dataSelectors=["PUBLIC_RELEASE"],
                failedFilePath=failedFilePath,
                saveInputFileListPath=saveInputFileListPath,
                pruneDocumentSize=pruneDocumentSize,
                validationLevel=schemaLevel,
            )
            okS = loadStatus(mw.getLoadStatus(), cfgOb, cachePath, readBackCheck=readBackCheck)

        if args.load_bird_ref:
            ok = mw.load(
                "bird",
                loadType=loadType,
                inputPathList=inputPathList,
                styleType=args.document_style,
                dataSelectors=["PUBLIC_RELEASE"],
                failedFilePath=failedFilePath,
                saveInputFileListPath=saveInputFileListPath,
                pruneDocumentSize=pruneDocumentSize,
                validationLevel=schemaLevel,
            )
            okS = loadStatus(mw.getLoadStatus(), cfgOb, cachePath, readBackCheck=readBackCheck)

        if args.load_bird_family_ref:
            ok = mw.load(
                "bird_family",
                loadType=loadType,
                inputPathList=inputPathList,
                styleType=args.document_style,
                dataSelectors=["BIRD_FAMILY_PUBLIC_RELEASE"],
                failedFilePath=failedFilePath,
                saveInputFileListPath=saveInputFileListPath,
                pruneDocumentSize=pruneDocumentSize,
                validationLevel=schemaLevel,
            )
            okS = loadStatus(mw.getLoadStatus(), cfgOb, cachePath, readBackCheck=readBackCheck)

        if args.load_entry_data:
            ok = mw.load(
                "pdbx",
                loadType=loadType,
                inputPathList=inputPathList,
                styleType=args.document_style,
                dataSelectors=["PUBLIC_RELEASE"],
                failedFilePath=failedFilePath,
                saveInputFileListPath=saveInputFileListPath,
                pruneDocumentSize=pruneDocumentSize,
                validationLevel=schemaLevel,
            )
            okS = loadStatus(mw.getLoadStatus(), cfgOb, cachePath, readBackCheck=readBackCheck)

        if args.load_pdbx_core:
            ok = mw.load(
                "pdbx_core",
                loadType=loadType,
                inputPathList=inputPathList,
                styleType=args.document_style,
                dataSelectors=["PUBLIC_RELEASE"],
                failedFilePath=failedFilePath,
                saveInputFileListPath=saveInputFileListPath,
                pruneDocumentSize=pruneDocumentSize,
                validationLevel=schemaLevel,
            )
            okS = loadStatus(mw.getLoadStatus(), cfgOb, cachePath, readBackCheck=readBackCheck)
        #
        if args.load_pdbx_core_merge:
            ok = mw.load(
                "pdbx_core",
                loadType=loadType,
                inputPathList=inputPathList,
                styleType=args.document_style,
                dataSelectors=["PUBLIC_RELEASE"],
                failedFilePath=failedFilePath,
                saveInputFileListPath=saveInputFileListPath,
                pruneDocumentSize=pruneDocumentSize,
                validationLevel=schemaLevel,
                mergeContentTypes=["vrpt"],
            )
            okS = loadStatus(mw.getLoadStatus(), cfgOb, cachePath, readBackCheck=readBackCheck)
        #
        if args.load_pdbx_core_entity:
            ok = mw.load(
                "pdbx_core",
                collectionLoadList=["pdbx_core_entity"],
                loadType=loadType,
                inputPathList=inputPathList,
                styleType=args.document_style,
                dataSelectors=["PUBLIC_RELEASE"],
                failedFilePath=failedFilePath,
                saveInputFileListPath=saveInputFileListPath,
                pruneDocumentSize=pruneDocumentSize,
                validationLevel=schemaLevel,
            )
            okS = loadStatus(mw.getLoadStatus(), cfgOb, cachePath, readBackCheck=readBackCheck)
        #
        if args.load_pdbx_core_entity_monomer:
            ok = mw.load(
                "pdbx_core",
                collectionLoadList=["pdbx_core_entity_monomer"],
                loadType=loadType,
                inputPathList=inputPathList,
                styleType=args.document_style,
                dataSelectors=["PUBLIC_RELEASE"],
                failedFilePath=failedFilePath,
                saveInputFileListPath=saveInputFileListPath,
                pruneDocumentSize=pruneDocumentSize,
                validationLevel=schemaLevel,
            )
            okS = loadStatus(mw.getLoadStatus(), cfgOb, cachePath, readBackCheck=readBackCheck)
        #
        if args.load_pdbx_core_entry:
            ok = mw.load(
                "pdbx_core",
                collectionLoadList=["pdbx_core_entry"],
                loadType=loadType,
                inputPathList=inputPathList,
                styleType=args.document_style,
                dataSelectors=["PUBLIC_RELEASE"],
                failedFilePath=failedFilePath,
                saveInputFileListPath=saveInputFileListPath,
                pruneDocumentSize=pruneDocumentSize,
                validationLevel=schemaLevel,
            )
            okS = loadStatus(mw.getLoadStatus(), cfgOb, cachePath, readBackCheck=readBackCheck)

        if args.load_pdbx_core_assembly:
            ok = mw.load(
                "pdbx_core",
                collectionLoadList=["pdbx_core_assembly"],
                loadType=loadType,
                inputPathList=inputPathList,
                styleType=args.document_style,
                dataSelectors=["PUBLIC_RELEASE"],
                failedFilePath=failedFilePath,
                saveInputFileListPath=saveInputFileListPath,
                pruneDocumentSize=pruneDocumentSize,
                validationLevel=schemaLevel,
            )
            okS = loadStatus(mw.getLoadStatus(), cfgOb, cachePath, readBackCheck=readBackCheck)

        if args.load_ihm_dev:
            ok = mw.load(
                "ihm_dev",
                loadType=loadType,
                inputPathList=inputPathList,
                styleType=args.document_style,
                dataSelectors=["PUBLIC_RELEASE"],
                failedFilePath=failedFilePath,
                saveInputFileListPath=saveInputFileListPath,
                pruneDocumentSize=pruneDocumentSize,
                validationLevel=schemaLevel,
            )
            okS = loadStatus(mw.getLoadStatus(), cfgOb, cachePath, readBackCheck=readBackCheck)
        #
        logger.info("Operation completed with status %r " % ok and okS)


if __name__ == "__main__":
    main()



##
# File:    DataTransformFactory.py
# Author:  J. Westbrook
# Date:    10-Apr-2018
#
#
# Updates:
#
#  4-Sep-2018 jdw add enumeration normalization (in progress)
# 13-Feb-2019 jdw limit translateXMLCharRefs() to categories/attributes
#                 configured with a transform filter.
#                 Refactor time critical sections of processRecords()
#                 to minimize costly functon calls for simple casts.
# 24-Mar-2019 jdw adjust null value filtering
##
"""
Factory for functional elements of the transformations between input data and
and loadable data using specifications from the schema map definition.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"


import collections
import logging
import re
from functools import reduce

import dateutil.parser
import pytz
from rcsb.db.utils.TextUtil import unescapeXmlCharRef

TrfValue = collections.namedtuple("TrfValue", "value, atId, origLength, isNull")

logger = logging.getLogger(__name__)


class DataTransformInfo(object):
    """Map transformation attribute filter names to data transformation filter implementations."""

    def __init__(self):
        # mapD {<external attribute filter name>: <implementation names>, ...}
        self.__mapD = {"STRIP_WS": "STRIP_WS", "TRANSLATE_XMLCHARREFS": "TRANSLATE_XMLCHARREFS"}

    def isImplemented(self, attributeFilter):
        ok = False
        try:
            if attributeFilter in self.__mapD:
                ok = True
        except Exception as e:
            logger.exception("Failing with %s", str(e))

        logger.debug("Filter %r status %r", attributeFilter, ok)
        return ok

    def getTransformFilterName(self, attributeFilter):
        try:
            return self.__mapD[attributeFilter]
        except Exception:
            pass
        return None


class DataTransformFactory(object):
    """Factory for functional elements of the transformations between input data and
    and loadable data using specifications from the schema map definition.

    input string value -> (Null Handling -> cast type ->)

    input string iterable value ->


    """

    def __init__(self, schemaDefAccessObj, filterType):
        self.__sD = schemaDefAccessObj
        self.__wsPattern = re.compile(r"\s+", flags=re.UNICODE | re.MULTILINE)
        logger.debug("filterType %r", filterType)
        logger.debug("Schema database %r", self.__sD.getDatabaseName())
        self.__transFlags = {}
        self.__transFlags["dropEmpty"] = "drop-empty-attributes" in filterType
        self.__transFlags["skipMaxWidth"] = "skip-max-width" in filterType
        self.__transFlags["assignDates"] = "assign-dates" in filterType
        self.__transFlags["convertIterables"] = "convert-iterables" in filterType
        self.__transFlags["normalizeEnums"] = "normalize-enums" in filterType
        self.__transFlags["translateXMLCharRefs"] = "translateXMLCharRefs" in filterType
        self.__transFlags["normalizeDates"] = True
        logger.debug("FLAGS settings are %r", self.__transFlags)
        #
        self.__wsPattern = re.compile(r"\s+", flags=re.UNICODE | re.MULTILINE)
        self.__dti = DataTransformInfo()
        self.__dT = self.__build()
        self.__nullValueD = {"string": "", "integer": r"\N", "float": r"\N", "date": r"\N", "datetime": r"\N"}

    def __build(self):
        """Internal method that stores transformations for each table so that these may
         be later repeatedly applied to a data stream.

        fD[tableId]'atFuncD'] -> {atId1: [f1,f2,.. ], atId2: [f1,f2,...], ...}
        """
        fD = {}
        for tableId in self.__sD.getSchemaIdList():
            tD = {}
            tObj = self.__sD.getSchemaObject(tableId)
            dt = DataTransform(tObj)
            aD = {}
            typeD = {}
            pureCastD = {}
            for atId in tObj.getAttributeIdList():
                if tObj.isOtherAttributeType(atId):
                    # skip attributes with no mapping correspondence
                    continue
                #
                aD[atId] = []
                #
                if self.__transFlags["convertIterables"] and tObj.isIterable(atId):
                    if tObj.isAttributeStringType(atId):
                        aD[atId].append(dt.castIterableString)
                        #
                        if self.__transFlags["translateXMLCharRefs"]:
                            aD[atId].append(dt.translateXMLCharRefsIt)
                    elif tObj.isAttributeIntegerType(atId):
                        aD[atId].append(dt.castIterableInteger)
                    elif tObj.isAttributeFloatType(atId):
                        aD[atId].append(dt.castIterableFloat)
                #
                elif tObj.isAttributeStringType(atId):
                    typeD[atId] = "string"
                    aD[atId].append(dt.castString)
                    if not self.__transFlags["skipMaxWidth"]:
                        aD[atId].append(dt.truncateString)
                    #
                    for ft in tObj.getAttributeFilterTypes(atId):
                        if self.__dti.getTransformFilterName(ft) == "STRIP_WS":
                            aD[atId].append(dt.stripWhiteSpace)
                        if self.__transFlags["translateXMLCharRefs"] and self.__dti.getTransformFilterName(ft) == "TRANSLATE_XMLCHARREFS":
                            aD[atId].append(dt.translateXMLCharRefs)
                    #
                    # if self.__transFlags['translateXMLCharRefs']:
                    #    aD[atId].append(dt.translateXMLCharRefs)
                elif tObj.isAttributeIntegerType(atId):
                    aD[atId].append(dt.castInteger)
                    typeD[atId] = "integer"
                elif tObj.isAttributeFloatType(atId):
                    aD[atId].append(dt.castFloat)
                    typeD[atId] = "float"
                elif self.__transFlags["assignDates"] and tObj.isAttributeDateType(atId):
                    typeD[atId] = "date"
                    aD[atId].append(dt.castDateToObj)
                elif self.__transFlags["normalizeDates"] and tObj.isAttributeDateType(atId):
                    if tObj.getAttributeType(atId).lower() == "datetime":
                        typeD[atId] = "datetime"
                        aD[atId].append(dt.castDateTimeToIsoDate)
                    elif tObj.getAttributeType(atId).lower() == "date":
                        aD[atId].append(dt.castDateToIsoDate)
                        typeD[atId] = "date"
                else:
                    aD[atId].append(dt.castString)
                    # typeD[tObj.getAttributeName(atId)] = 'string'
                    typeD[atId] = "string"
                #
                if self.__transFlags["normalizeEnums"] and tObj.isEnumerated(atId):
                    logger.debug("Normalizing enums for %s %s", tableId, atId)
                    aD[atId].append(dt.normalizeEnum)
                #
                if len(aD[atId]) == 1 and atId in typeD and typeD[atId] in ["string", "float", "integer"]:
                    pureCastD[tObj.getAttributeName(atId)] = typeD[atId]
            # Transformation functions keyed by attribute 'name'
            tD["atIdD"] = tObj.getMapAttributeIdDict()
            tD["atNameD"] = tObj.getMapAttributeNameDict()
            tD["atNullValues"] = tObj.getAppNullValueDict()
            tD["atFuncD"] = {tD["atIdD"][k]: v for k, v in aD.items()}
            tD["pureCast"] = pureCastD
            #
            fD[tableId] = tD
        ##
        return fD

    def get(self, tableId):
        try:
            return self.__dT[tableId]
        except Exception as e:
            logger.error("Missing table %r with error %s", tableId, str(e))
        return {}

    def processRecord(self, tableId, row, attributeNameList, containerName=None):
        """
        Input row data (list) ordered according to the input attribute names list.

        Processing respects various null handling policies.


        return   d[atId]=rowdata for the input row list

        """
        # get the transform object for the current table
        #
        # Avoiding method call ...
        # dT = self.get(tableId)
        dT = self.__dT[tableId] if tableId in self.__dT else {}
        #
        atName = None
        try:
            dD = {} if self.__transFlags["dropEmpty"] else {k: v for k, v in dT["atNullValues"].items()}
            for ii, atName in enumerate(attributeNameList):
                if atName not in dT["atNameD"]:
                    continue
                #
                # nullFlag = False if row[ii] else True
                nullFlag = False if row[ii] is not None else True
                #
                # Incorporate pure casts into this loop for performance -
                if atName in dT["pureCast"]:
                    if nullFlag and self.__transFlags["dropEmpty"]:
                        continue
                    if (row[ii] == "?") or (row[ii] == ".") or (row[ii]) == "":
                        if self.__transFlags["dropEmpty"]:
                            continue
                        else:
                            dD[dT["atNameD"][atName]] = self.__nullValueD[dT["pureCast"][atName]]
                            continue
                    else:
                        if dT["pureCast"][atName] == "string":
                            dD[dT["atNameD"][atName]] = row[ii]
                        elif dT["pureCast"][atName] == "integer":
                            dD[dT["atNameD"][atName]] = int(row[ii])
                        elif dT["pureCast"][atName] == "float":
                            dD[dT["atNameD"][atName]] = float(row[ii])
                        continue
                else:
                    # Apply list of functions on an initial value (i.e. TrfValue for the ii(th) element of the row.
                    vT = reduce(lambda x, y: y(x), dT["atFuncD"][atName], TrfValue(row[ii], dT["atNameD"][atName], 0, nullFlag))
                    if self.__transFlags["dropEmpty"] and vT.isNull:
                        continue
                    dD[dT["atNameD"][atName]] = vT.value
        except Exception as e:
            logger.error("Failing for %r table %s atName %s with %s", containerName, tableId, atName, str(e))
            # logger.exception("Failing for %r table %r atName %r with %s", containerName, tableId, atName, str(e))

        return dD


class DataTransform(object):
    """Factory for functional elements of the transformations between input data and
    and loadable data using specifications from the schema map definition.

    input string value -> (Null Handling -> cast type ->)

    input string iterable value ->
    """

    def __init__(self, tObj):
        #
        self.__wsPattern = re.compile(r"\s+", flags=re.UNICODE | re.MULTILINE)

        # SchemaDef Table Object -
        #
        #
        # Set null / missing value place holders
        self.__nullValueString = ""
        self.__nullValueOther = r"\N"
        self.__tObj = tObj
        #

    def normalizeEnum(self, trfTup):
        """
        Return:  TrfValue tuple
        """
        if trfTup.isNull:
            return trfTup
        # origLength = len(trfTup.value)
        # if ((origLength == 0) or (trfTup.value == '?') or (trfTup.value == '.')):
        #    return TrfValue(self.__nullValueString, trfTup.atId, origLength, True)
        if trfTup.value and isinstance(trfTup.value, (list,)):
            tL = [self.__tObj.normalizeEnum(trfTup.atId, t) for t in trfTup.value]
            nVal = tL
        else:
            nVal = self.__tObj.normalizeEnum(trfTup.atId, trfTup.value)
        # logger.info("Normalizing %r %r - normalized %r", trfTup.atId, trfTup.value, nVal)
        return TrfValue(nVal, trfTup.atId, trfTup.origLength, False)

    def castStringX(self, trfTup):
        """
        Return:  TrfValue tuple
        """
        if trfTup.isNull:
            return trfTup
        origLength = len(trfTup.value)
        if (origLength == 0) or (trfTup.value == "?") or (trfTup.value == "."):
            return TrfValue(self.__nullValueString, trfTup.atId, origLength, True)
        return TrfValue(trfTup.value, trfTup.atId, origLength, False)

    def castString(self, trfTup):
        """
        Return:  TrfValue tuple
        """
        if trfTup.isNull:
            return trfTup
        if (trfTup.value == "?") or (trfTup.value == ".") or (not trfTup.value):
            return TrfValue(self.__nullValueString, trfTup.atId, trfTup.origLength, True)
        return TrfValue(trfTup.value, trfTup.atId, trfTup.origLength, False)

    def castIterableString(self, trfTup):
        """
        Return:  TrfValue tuple
        """
        if trfTup.isNull:
            return trfTup
        if (trfTup.value == "?") or (trfTup.value == ".") or (not trfTup.value):
            return TrfValue(self.__nullValueString, trfTup.atId, trfTup.origLength, True)
        vL = [v.strip() for v in trfTup.value.split(self.__tObj.getIterableSeparator(trfTup.atId))]
        return TrfValue(vL, trfTup.atId, trfTup.origLength, False)

    def castInteger(self, trfTup):
        """
        Return:  TrfValue tuple
        """
        if trfTup.isNull:
            return trfTup
        if (trfTup.value == "?") or (trfTup.value == ".") or (trfTup.value is None) or (trfTup.value == ""):
            return TrfValue(self.__nullValueOther, trfTup.atId, trfTup.origLength, True)
        return TrfValue(int(trfTup.value), trfTup.atId, trfTup.origLength, False)

    def castIterableInteger(self, trfTup):
        """
        Return:  TrfValue tuple
        """
        if trfTup.isNull:
            return trfTup
        if (trfTup.value == "?") or (trfTup.value == ".") or (trfTup.value is None) or (trfTup.value == ""):
            return TrfValue(self.__nullValueOther, trfTup.atId, trfTup.origLength, True)
        # vL = [int(v.strip()) for v in str(trfTup.value).split(self.__tObj.getIterableSeparator(trfTup.atId))]
        vL = [int(v.strip()) if v.strip() not in [".", "?"] else None for v in str(trfTup.value).split(self.__tObj.getIterableSeparator(trfTup.atId))]
        return TrfValue(vL, trfTup.atId, trfTup.origLength, False)

    def castFloat(self, trfTup):
        """
        Return:  TrfValue tuple
        """
        if trfTup.isNull:
            return trfTup
        if (trfTup.value == "?") or (trfTup.value == ".") or (trfTup.value is None) or (trfTup.value == ""):
            return TrfValue(self.__nullValueOther, trfTup.atId, trfTup.origLength, True)
        return TrfValue(float(trfTup.value), trfTup.atId, trfTup.origLength, False)

    def castIterableFloat(self, trfTup):
        """
        Return:  TrfValue tuple
        """
        # logger.info(">> atId %r value %r delimiter %r", trfTup.atId, trfTup.value, self.__tObj.getIterableSeparator(trfTup.atId))
        if trfTup.isNull:
            return trfTup
        if (trfTup.value == "?") or (trfTup.value == ".") or (trfTup.value is None) or (trfTup.value == ""):
            return TrfValue(self.__nullValueOther, trfTup.atId, trfTup.origLength, True)
        # vL = [float(v.strip()) for v in str(trfTup.value).split(self.__tObj.getIterableSeparator(trfTup.atId))]
        vL = [float(v.strip()) if v.strip() not in [".", "?"] else None for v in str(trfTup.value).split(self.__tObj.getIterableSeparator(trfTup.atId))]
        return TrfValue(vL, trfTup.atId, trfTup.origLength, False)

    def castDateToObj(self, trfTup):
        """Cast the input date (optional time) string (yyyy-mm-dd:hh::mm:ss) to a Python DateTime object -

        Return:  TrfValue tuple
        """
        if trfTup.isNull:
            return trfTup
        origLength = len(trfTup.value)
        if (origLength == 0) or (trfTup.value == "?") or (trfTup.value == "."):
            return TrfValue(self.__nullValueOther, trfTup.atId, origLength, True)
        tv = trfTup.value.replace(":", " ", 1)
        return TrfValue(dateutil.parser.parse(tv), trfTup.atId, origLength, False)

    def castDateTimeToIsoDate(self, trfTup):
        """Cast the input date (optional time) string (yyyy-mm-dd:hh::mm:ss) to a Python DateTime object -

        Return:  TrfValue tuple
        """
        if trfTup.isNull:
            return trfTup
        origLength = len(trfTup.value)
        if (origLength == 0) or (trfTup.value == "?") or (trfTup.value == "."):
            return TrfValue(self.__nullValueOther, trfTup.atId, origLength, True)
        tv = trfTup.value.replace(":", " ", 1)
        tS = dateutil.parser.parse(tv).replace(tzinfo=pytz.UTC).isoformat()

        return TrfValue(tS, trfTup.atId, origLength, False)

    def castDateToIsoDate(self, trfTup):
        """Cast the input date (optional time) string (yyyy-mm-dd:hh::mm:ss) to a Python DateTime object -

        Return:  TrfValue tuple
        """
        if trfTup.isNull:
            return trfTup
        origLength = len(trfTup.value)
        if (origLength == 0) or (trfTup.value == "?") or (trfTup.value == "."):
            return TrfValue(self.__nullValueOther, trfTup.atId, origLength, True)
        tv = trfTup.value.replace(":", " ", 1)
        tS = dateutil.parser.parse(tv).isoformat()

        return TrfValue(tS[:10], trfTup.atId, origLength, False)

    def castDateToString(self, trfTup):
        """Cast the input date (optional time) string (yyyy-mm-dd:hh::mm:ss) as a string unchanged -

        Return:  TrfValue tuple
        """
        if trfTup.isNull:
            return trfTup
        origLength = len(trfTup.value)
        if (origLength == 0) or (trfTup.value == "?") or (trfTup.value == "."):
            return TrfValue(self.__nullValueOther, trfTup.atId, origLength, True)
        return TrfValue(trfTup.value, trfTup.atId, origLength, False)

    def stripWhiteSpace(self, trfTup):
        """Remove all white space from the input value.

        Return:  ReturnValue tuple
        """
        if trfTup.isNull:
            return trfTup
        value = self.__wsPattern.sub("", trfTup.value)
        return TrfValue(value, trfTup.atId, trfTup.origLength, False)

    def truncateString(self, trfTup):
        """Truncate string value to maximum length setting.

        Return:  ReturnValue tuple
        """
        if trfTup.isNull:
            return trfTup
        return TrfValue(trfTup.value[: self.__tObj.getAttributeWidth(trfTup.atId)], trfTup.atId, trfTup.origLength, False)

    def translateXMLCharRefs(self, trfTup):
        """Convert XML Character references to unicode.

        Return:  ReturnValue tuple
        """
        if trfTup.isNull:
            return trfTup
        return TrfValue(unescapeXmlCharRef(trfTup.value), trfTup.atId, trfTup.origLength, False)

    def translateXMLCharRefsIt(self, trfTup):
        """Convert XML Character references to unicode.

        Return:  ReturnValue tuple
        """
        if trfTup.isNull:
            return trfTup
        #
        vL = [unescapeXmlCharRef(v) for v in trfTup.value]
        return TrfValue(vL, trfTup.atId, trfTup.origLength, False)
